"""Tests suite for the data-validation-framework package."""
import re
from pathlib import Path


def check_files_exist(directory, pattern_list):
    """Check that a list of file patterns exist in a given directory."""
    existing_files = [str(i.relative_to(directory)) for i in sorted(Path(directory).rglob("*"))]
    try:
        assert len(existing_files) == len(pattern_list)
        for path, pattern in zip(existing_files, pattern_list):
            assert re.match(str(pattern), path)
    except Exception as exc:
        formatted_existing_files = "\n\t\t".join(existing_files)
        formatted_pattern_list = "\n\t\t".join(pattern_list)
        raise RuntimeError(
            "Error when checking the files.\n\t"
            f"The directory is: {directory}\n"
            "\t"
            f"The found files are:\n\t\t{formatted_existing_files}\n"
            "\t"
            f"The patterns are:\n\t\t{formatted_pattern_list}"
        ) from exc
    return True



"""Configuration for the pytest test suite."""
# pylint: disable=missing-function-docstring
import os
from pathlib import Path

import pytest

from data_validation_framework.target import OutputLocalTarget

DATA = Path(__file__).parent / "data"


@pytest.fixture()
def tmp_working_dir(tmp_path):
    """Change working directory before a test and change it back when the test is finished."""
    cwd = os.getcwd()
    os.chdir(tmp_path)
    yield tmp_path
    os.chdir(cwd)


@pytest.fixture()
def data_dir():
    """Path to the directory where the data are stored."""
    return DATA


@pytest.fixture(autouse=True)
def reset_target_prefix(tmpdir):
    """Automatically set the default prefix to the current test directory."""
    OutputLocalTarget.set_default_prefix(tmpdir)
    yield
    OutputLocalTarget.set_default_prefix(None)



"""Test the data_validation_framework.rst_tools module."""
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=redefined-outer-name
# pylint: disable=use-implicit-booleaness-not-comparison
import pytest

from data_validation_framework import rst_tools


def test_init():
    """Test constructor."""
    rst = rst_tools.RstFile("test_file")
    assert rst.file_path == "test_file"
    assert rst.mode == "w"
    assert rst._data == []  # pylint: disable=protected-access

    with pytest.raises(ValueError, match="The 'file_path' argument must be a not empty string."):
        rst_tools.RstFile("")

    with pytest.raises(ValueError, match="The 'file_path' argument must be a not empty string."):
        rst_tools.RstFile(None)


@pytest.fixture
def empty_RstFile(tmpdir):
    """Create and empty RstFile object."""
    filename = tmpdir / "test_file.rst"
    return rst_tools.RstFile(str(filename))


def test_add(empty_RstFile):
    """Test add method."""
    empty_RstFile.add("test")
    empty_RstFile.add("test", indent=2)
    empty_RstFile.add("long\nline\nwith\nnewlines", indent=2)
    assert empty_RstFile.get_buffer() == ["test", "  test", "  long\nline\nwith\nnewlines"]


def test_new_line(empty_RstFile):
    """Test new_line method."""
    empty_RstFile.newline()
    empty_RstFile.newline(3)
    assert empty_RstFile.get_buffer() == ["\n"] * 4

    with pytest.raises(ValueError, match="Count of newlines must be a positive int."):
        empty_RstFile.newline(-1)

    with pytest.raises(ValueError, match="Count of newlines must be a positive int."):
        empty_RstFile.newline(None)


def test_clear(empty_RstFile):
    """Test clear method."""
    empty_RstFile.add("test")
    assert empty_RstFile.get_buffer() == ["test"]
    empty_RstFile.clear()
    assert empty_RstFile.get_buffer() == []


def test_str_formatting():
    """Test formatting methods."""
    assert rst_tools.RstFile.bold("test") == "**test**"
    assert rst_tools.RstFile.emph("test") == "*test*"
    assert rst_tools.RstFile.pre("test") == "``test``"
    assert (
        rst_tools.RstFile.inline_link("the text", "http://the.link")
        == "`the text <http://the.link>`_"
    )
    # pylint: disable=protected-access
    assert rst_tools.RstFile._indent("test") == "test"
    assert rst_tools.RstFile._indent("test", 2) == "  test"
    assert rst_tools.RstFile._indent(["test1", "test2"], 2) == ["  test1", "  test2"]


def test_headings(empty_RstFile):
    """Test heading methods."""
    empty_RstFile.title("The title")
    empty_RstFile.h1("The heading 1")
    empty_RstFile.h2("The heading 2")
    empty_RstFile.h3("The heading 3")
    empty_RstFile.h4("The heading 4")
    empty_RstFile.h5("The heading 5")
    empty_RstFile.h6("The heading 6")
    empty_RstFile.heading_level(3, "The heading 6")

    assert empty_RstFile.get_buffer() == [
        "=========",
        "The title",
        "=========",
        "\n",
        "\n",
        "The heading 1",
        "=============",
        "\n",
        "\n",
        "The heading 2",
        "-------------",
        "\n",
        "\n",
        "The heading 3",
        "~~~~~~~~~~~~~",
        "\n",
        "\n",
        "The heading 4",
        "+++++++++++++",
        "\n",
        "\n",
        "The heading 5",
        "^^^^^^^^^^^^^",
        "\n",
        "\n",
        "The heading 6",
        ";;;;;;;;;;;;;",
        "\n",
        "\n",
        "The heading 6",
        "~~~~~~~~~~~~~",
        "\n",
    ]


def test_toc_tree(empty_RstFile):
    """Test toc_tree method."""
    # Default
    empty_RstFile.toc_tree()

    # With max depth
    empty_RstFile.clear()
    empty_RstFile.toc_tree(maxdepth=3)
    assert empty_RstFile.get_buffer() == [
        "\n",
        ".. toctree::",
        "   :maxdepth: 3",
        "   :numbered:",
        "\n",
    ]

    # With not numbered
    empty_RstFile.clear()
    empty_RstFile.toc_tree(numbered=False)
    assert empty_RstFile.get_buffer() == ["\n", ".. toctree::", "\n"]

    # With indent
    empty_RstFile.clear()
    empty_RstFile.toc_tree(indent=2)
    assert empty_RstFile.get_buffer() == ["\n", "  .. toctree::", "     :numbered:", "\n"]

    # With contents
    empty_RstFile.clear()
    empty_RstFile.toc_tree(toc_type="contents")
    assert empty_RstFile.get_buffer() == ["\n", ".. contents::", "\n"]

    # With bad toc type
    empty_RstFile.clear()
    with pytest.raises(ValueError, match="Unknown TOC tree type"):
        empty_RstFile.toc_tree(toc_type="bad type")


def test_code_block(empty_RstFile):
    """Test code_block methods."""
    # Default
    empty_RstFile.code_block()
    assert empty_RstFile.get_buffer() == ["\n", ".. code-block:: ", "\n"]

    # With language
    empty_RstFile.clear()
    empty_RstFile.code_block(language="Python")
    assert empty_RstFile.get_buffer() == ["\n", ".. code-block:: Python", "\n"]


def test_exception(empty_RstFile):
    """Test exception methods."""
    exception_str = (
        "\t\u001b[0;35mException\u001b[0m "
        "\u001b[0;36m172.18.0.2\u001b[0m "
        "\u001b[0;36m/file/path.test\u001b[0m"
    )

    # Default
    empty_RstFile.exception()
    assert empty_RstFile.get_buffer() == ["\n", "", "\n"]

    # With ANSI escape sequences
    empty_RstFile.clear()
    empty_RstFile.exception(exception_str)
    assert empty_RstFile.get_buffer() == ["\n", "\tException 172.18.0.2 /file/path.test", "\n"]

    # With width
    empty_RstFile.clear()
    empty_RstFile.exception(exception_str, width=15)
    assert empty_RstFile.get_buffer() == ["\n", "Exception", "172.18.0.2", "/file/path.test", "\n"]


def test_list(empty_RstFile):
    """Test list methods."""
    list_str = [
        "first",
        "second",
        "third",
    ]

    # Default
    empty_RstFile.list(list_str)
    assert empty_RstFile.get_buffer() == ["* first", "* second", "* third"]

    # With custom bullet
    empty_RstFile.clear()
    empty_RstFile.list(list_str, bullet="-")
    assert empty_RstFile.get_buffer() == ["- first", "- second", "- third"]

    # With empty list
    empty_RstFile.clear()
    empty_RstFile.list([])
    assert empty_RstFile.get_buffer() == []

    # With width
    empty_RstFile.clear()
    empty_RstFile.list([" ".join([i] * 4) for i in list_str], width=15)
    assert empty_RstFile.get_buffer() == [
        "* first first\n  first first",
        "* second second\n  second second",
        "* third third\n  third third",
    ]

    # With width and indent
    empty_RstFile.clear()
    empty_RstFile.list([" ".join([i] * 4) for i in list_str], width=10, indent=2)
    assert empty_RstFile.get_buffer() == [
        "  * first\n    first\n    first\n    first",
        "  * second\n    second\n    second\n    second",
        "  * third\n    third\n    third\n    third",
    ]


def test_write(empty_RstFile):
    """Test write methods."""
    empty_RstFile.title("The title")
    empty_RstFile.add("Some text")
    empty_RstFile.write()

    with open(empty_RstFile.file_path, encoding="utf-8") as res_file:
        res = res_file.read()

    assert res == ("=========\n" "The title\n" "=========\n" "\n" "\n" "Some text\n")



"""data-validation-framework package.

Simple framework to create data validation workflows.
"""
import importlib.metadata

from data_validation_framework import result  # noqa
from data_validation_framework import task  # noqa

__version__ = importlib.metadata.version("data-validation-framework")



"""Specific tasks."""
import logging
import sys
import time
import traceback
import warnings
from functools import partial
from pathlib import Path

import luigi
import pandas as pd
from luigi.parameter import ChoiceParameter
from luigi.parameter import DictParameter
from luigi.parameter import OptionalIntParameter
from luigi.parameter import OptionalPathParameter
from luigi.parameter import OptionalStrParameter
from luigi.task import flatten as task_flatten
from luigi_tools.parameter import BoolParameter
from luigi_tools.parameter import OptionalBoolParameter
from luigi_tools.task import LogTargetMixin
from luigi_tools.task import RerunMixin
from numpy import VisibleDeprecationWarning

from data_validation_framework.report import make_report
from data_validation_framework.result import ValidationResult
from data_validation_framework.result import ValidationResultSet
from data_validation_framework.target import DataDirectoryTarget
from data_validation_framework.target import ReportTarget
from data_validation_framework.target import TaggedOutputLocalTarget
from data_validation_framework.util import apply_to_df

L = logging.getLogger(__name__)
INDEX_LABEL = "__index_label__"
SKIP_COMMENT = "Skipped by user."


class ValidationError(Exception):
    """Exception raised if the validation could not be performed properly."""


class TagResultOutputMixin:
    """Initialize target prefixes and optionally add a tag to the resut directory.

    .. warning::

        If this mixin is used alongside the :class:`luigi_tools.task.RerunMixin`, then it can have
        two different behaviors:

        * if placed on the right side of the `RerunMixin`, a new tag is created and thus rerun does
          not do anything.
        * if placed on the left side of the `RerunMixin`, the targets from the tagged directory
          without conflict are removed then a new tagged directory with conflict is created.
    """

    tag_output = BoolParameter(
        default=False, description=":bool: Add a tag suffix to the output directory."
    )
    result_path = OptionalPathParameter(
        default=None, description=":str: Path to the global output directory."
    )
    target_cls = TaggedOutputLocalTarget

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        def num_tag(path, num):
            return path.with_name(f"{path.name}_{num}")

        if self.tag_output and self.result_path is not None:
            path = Path(f"{self.result_path}_{time.strftime('%Y%m%d-%Hh%Mm%Ss')}")
            tagged_output = path
            num = 0
            while tagged_output.exists():
                L.debug("Tagged output '%s' already exists", tagged_output)
                num += 1
                tagged_output = num_tag(path, num)

            if num > 0 and RerunMixin in self.__class__.mro() and self.rerun:
                warnings.warn(
                    "Using 'rerun' with conflicting tag output results in creating a new tag or "
                    "removing the untagged result directory (depending on the inheritance order).",
                )

            L.info("Tagged output is: %s", tagged_output)
            self.result_path = tagged_output


class InputParameters:
    """A helper to build task inputs.

    This class should only be used to build the input dictionaries:

    .. code-block:: python

        class Task(ChildBaseValidationTask):
            data_dir = luigi.Parameter(default='my_data')
            validation_function = staticmethod(my_validation_function)
            output_columns = {'my_new_col': None}

            def inputs(self):
                return {
                    PreviousTask: InputParameters(
                        {"input_data", "input_data"},
                        kwarg_1="value_1",
                        kwarg_2="value_2"
                    )
                }

    Args:
        col_mapping (dict): The column mapping.

    Keyword Args:
        kwargs: All the keyword arguments passed to the constructor.

    .. warning:: The keyword arguments will always override the arguments from the config file.
    """

    def __init__(self, col_mapping, **kwargs):
        self.col_mapping = col_mapping
        self.kwargs = kwargs


class BaseValidationTask(
    LogTargetMixin,
    RerunMixin,
    TagResultOutputMixin,
    luigi.Task,
):
    """Base luigi task used for validation steps.

    .. warning::
        Only use child of this class such as :class:`ElementValidationTask` or
        :class:`SetValidationTask`.

    To provide a validation function, use it as:

    .. code-block:: python

        class Task(ChildBaseValidationTask):
            data_dir = luigi.Parameter(default='my_data')
            validation_function = staticmethod(my_validation_function)
            output_columns = {'my_new_col': None}

            def inputs(self):
                return {PreviousTask: {"input_data", "input_data"}

    .. note::

        The ``dataset_df`` and ``result_path`` attributes are passed to the requirements when they
        have ``None`` values for these attributes.

    """

    # I/O Parameters
    dataset_df = OptionalPathParameter(
        description=":str: Path to the input dataset.",
        default=None,
        exists=True,
    )
    input_index_col = OptionalStrParameter(
        default=None, description=":str: Name of the column used as index."
    )
    data_dir = OptionalStrParameter(
        default="data",
        description=(
            ":str: name of folder to store addittional files created by a task (the provided "
            "validation function must take this as argument)."
        ),
    )

    # Naming Parameters
    custom_task_name = OptionalStrParameter(
        default=None,
        significant=False,
        description=(
            ":str: An optional custom name given to the task (the class name is used if not "
            "provided)."
        ),
    )

    # User attributes
    output_columns = None
    """dict: A dict with names as keys and empty values as values for new columns created by the
    current task.
    """

    gather_inputs = False
    """bool: Gather the inputs in the :class:`pandas.DataFrame` given to the
    :meth:`validation_function`.
    """

    # Internal attributes
    nb_total = None
    """int: Total number of processed elements."""
    nb_valid = None
    """int: Total number of valid elements."""
    results = None
    """pandas.DataFrame: The results of the :meth:`validation_function`."""

    nb_processes = OptionalIntParameter(
        default=None,
        description=":int: The number of parallel processes to use.",
        significant=False,
    )

    redirect_stdout = OptionalBoolParameter(
        default=None,
        parsing=luigi.BoolParameter.EXPLICIT_PARSING,
        description=(
            ":bool: Capture stdout from the validation function to make it work with progress "
            "bar. Disable it if you want to use PDB inside the validation function."
        ),
        significant=False,
    )

    def __init__(self, *args, **kwargs):
        warnings.filterwarnings("ignore", module="numpy", category=VisibleDeprecationWarning)
        warnings.filterwarnings("ignore", module="luigi.task", category=DeprecationWarning)
        event_handler = super().event_handler

        # pylint: disable=unused-variable
        @event_handler(luigi.Event.SUCCESS)
        def success_summary(self):
            """Summary report of the task."""
            L.info("==========================================")
            task_summary = f"SUMMARY {self.task_name}: {self.nb_valid} / {self.nb_total} passed"
            L.info(task_summary)
            L.info("==========================================")

        super().__init__(*args, **kwargs)

        self._report_cols = list(ValidationResultSet.out_cols.keys())

    def inputs(self):
        """Information about required input data.

        This method can be overridden and should return a dict of one of the following forms:

        .. code-block:: python

            {<task_name>: {"<input_column_name>": "<current_column_name>"}}

        .. code-block:: python

            {
                <task_name>: (
                    {"<input_column_name>": "<current_column_name>"},
                    {
                        "<kwarg_1>": "<value_1>",
                        "<kwarg_2>": "<value_2>",
                    }
                )
            }

        .. code-block:: python

            {
                <task_name>: InputParameters(
                    {"<input_column_name>": "<current_column_name>"},
                    kwarg_1="<value_1>",
                    kwarg_2="<value_2>",
                )
            }

        where:
            - ``<task_name>`` is the name of the required task,
            - ``<input_column_name>`` is the name of the column we need from the report of
              task_name,
            - ``<current_column_name>`` is the name of the same column in the current task.
            - ``<kwarg_*>`` is the name of a keyword argument passed to the constructor of the
              required task.
        """
        return None

    def args(self):
        """List of addititonal arguments to provide to :meth:`validation_function`.

        This method can be overridden to pass these arguments to the :meth:`validation_function`
        and to the :meth:`pre_process` and :meth:`post_process` methods.
        """
        return []

    def kwargs(self):
        """Dict of addititonal keyword arguments to provide to :meth:`validation_function`.

        This method can be overridden to pass these keyword arguments to the
        :meth:`validation_function` and to the :meth:`pre_process` and
        :meth:`post_process` methods.
        """
        return {}

    def read_dataset(self):
        """Import the dataset to a :class:`pandas.DataFrame`.

        This method can be overridden to load custom data (e.g. GeoDataFrame, etc.).
        The dataset should always be loaded from the path given by `self.dataset_df`.
        """
        return pd.read_csv(self.dataset_df, index_col=self.input_index_col)

    def pre_process(self, df, args, kwargs):
        """Method executed before applying the external function."""

    def post_process(self, df, args, kwargs):
        """Method executed after applying the external function."""

    def processed_inputs(self):
        """Process the inputs to automatically propagate the values from the workflow."""
        inputs = self.inputs()  # pylint: disable=assignment-from-none

        if not inputs:
            return None

        def default_kwargs(self):
            return {
                "dataset_df": self.dataset_df,
                "input_index_col": self.input_index_col,
                "result_path": self.result_path,
                "nb_processes": self.nb_processes,
                "redirect_stdout": self.redirect_stdout,
            }

        formatted_inputs = {}
        for task, v in inputs.items():
            if isinstance(v, dict):
                col_mapping = v
                specific_kwargs = {}
            elif isinstance(v, InputParameters):
                col_mapping = v.col_mapping
                specific_kwargs = v.kwargs
            else:
                try:
                    specific_kwargs, col_mapping = v
                except ValueError as exc:
                    raise ValueError(
                        "The input values should either be a dict containing the column mapping "
                        "or a tuple with a dict containing the keyword arguments as first element "
                        "and a dict containing the column mapping as second element."
                    ) from exc
            kwargs = default_kwargs(self)
            base_kwargs = {
                key: value
                for key, value in task.get_param_values(task.get_params(), [], specific_kwargs)
                if value is not None
            }
            kwargs.update(base_kwargs)
            formatted_inputs[task(**kwargs)] = col_mapping
        return formatted_inputs

    def requires(self):
        """Process the inputs to generate the requirements."""
        requires = list(self.processed_inputs() or [])
        return requires + task_flatten(self.extra_requires())

    def extra_requires(self):
        """Requirements that should not be considered as validation tasks."""
        return []

    def extra_input(self):
        """Targets of the tasks given to extra_requires()."""
        return luigi.task.getpaths(self.extra_requires())

    @staticmethod
    def _rename_cols(df):
        if df.columns.nlevels > 1:
            df.columns = pd.MultiIndex.from_tuples(
                [tuple("" if "Unnamed: " in j else j for j in i) for i in df.columns.tolist()]
            )
        return df

    def _process(self, df, *args, **kwargs):  # pragma: no cover
        raise NotImplementedError("Please do not use the BaseValidationTask but deriving class.")

    @staticmethod
    def validation_function(*args, **kwargs):  # pragma: no cover
        """The validation function to apply to the current data set."""
        raise ValueError(
            "You must set the 'validation_function' attribute of the class to a reference to a "
            "function."
        )

    def __getattribute__(self, name):
        """Unbound the validation function if it is bounded to the current instance."""
        if name == "validation_function":
            func = super().__getattribute__(name)
            if hasattr(func, "__func__"):
                return func.__func__
            return func

        return super().__getattribute__(name)

    @staticmethod
    def check_inputs(inputs):
        """Check a given dict of inputs."""
        if not inputs:
            return False

        # Check column names conflicts
        forbidden_cols = ValidationResultSet.out_cols.keys()
        col_names = {}
        for t, col_mapping in inputs.items():
            for new_col in col_mapping.values():
                checked = col_names.get(new_col)
                if checked is not None:
                    raise ValueError(
                        f"The destination column '{new_col}' of the tasks {t} and {checked} are "
                        "conflicting."
                    )
                if new_col in forbidden_cols:
                    raise ValueError(
                        f"The destination column of the task {t} can not be one of "
                        f"{list(forbidden_cols)}."
                    )
                col_names[new_col] = t

        # Check if column mapping was found
        if not col_names:
            return False
        return True

    @staticmethod
    def filter_columns(all_dfs, mappings):
        """Filter the columns of the given DataFrames according to a given mapping."""
        filtered_dfs = []
        for t, df in all_dfs.items():
            col_mapping = mappings.get(t, {})

            # Pick and rename columns
            try:
                filtered_dfs.append(df[col_mapping.keys()].rename(columns=col_mapping))
            except KeyError as e:
                missing_cols = sorted(set(col_mapping.keys()) - set(df.columns))
                raise KeyError(
                    f"The columns {missing_cols} are missing from the output columns of {t}."
                ) from e

        return filtered_dfs

    def _get_dataset(self):
        """Get the dataset."""
        if self.dataset_df is not None:
            L.info("Input dataset: %s", Path(self.dataset_df).resolve())
            new_df = self.read_dataset()
            duplicated_index = new_df.index.duplicated()
            if duplicated_index.any():
                raise IndexError(
                    "The following index values are duplicated: "
                    f"{sorted(new_df.index.to_series().loc[duplicated_index])}"
                )
        else:
            new_df = pd.DataFrame()

        return new_df

    def _join_inputs(self, new_df):
        """Get the inputs and join them to the dataset."""
        if self.processed_inputs():
            # Check inputs
            with_mapping = self.check_inputs(
                self.processed_inputs()
            )  # pylint: disable=not-callable

            # Get the input targets and their DataFrames
            all_inputs = {
                i: [
                    target
                    for target in task_flatten(i.output())
                    if isinstance(target, ReportTarget)
                ]
                for i in task_flatten(self.processed_inputs().keys())
            }
            all_report_paths = {
                t: [r.path for r in reports][0] for t, reports in all_inputs.items()
            }
            L.debug("Importing the following reports: %s", all_report_paths)
            all_dfs = {
                task_obj: self._rename_cols(
                    pd.read_csv(path, index_col=INDEX_LABEL).rename_axis(index="index")
                )
                for task_obj, path in all_report_paths.items()
            }

            # Compute the global is_valid column
            are_valid = pd.concat([i["is_valid"] for i in all_dfs.values()], axis=1)
            is_valid = are_valid.all(axis=1)
            new_df["is_valid"] = is_valid

            # Filter columns in the DataFrames
            if with_mapping:
                # pylint: disable=not-callable
                filtered_dfs = self.filter_columns(all_dfs, self.processed_inputs())

                # Concatenate all DataFrames
                filtered_df = pd.concat(filtered_dfs, axis=1)

                # Erase columns
                new_df[filtered_df.columns] = filtered_df

                # Add filtered columns to the report ones
                self._report_cols.extend(
                    [col for col in filtered_df.columns if col not in self._report_cols]
                )

            if self.gather_inputs:
                # Use column MultiIndex on each input DataFrame in order to gather them
                for task_obj, df in all_dfs.items():
                    df.columns = pd.MultiIndex.from_tuples(
                        [(task_obj.task_name, i) for i in df.columns.values]
                    )

                # Concatenate all DataFrames
                gathered = pd.concat(all_dfs.values(), axis=1)
                gathered.columns = gathered.columns.to_flat_index()

                # Add the gathered DataFrame to the new DataFrame
                new_df = new_df.join(gathered)

                # Add gathered columns to the report ones
                self._report_cols.extend(
                    [col for col in gathered.columns if col not in self._report_cols]
                )

            # Clean inconsistent indexes
            null_index = new_df["is_valid"].isnull()
            if null_index.any():
                L.warning(
                    "The following inconsistent indexes between the dataset and the inputs are "
                    "ignored: %s",
                    null_index.loc[null_index].index.tolist(),
                )
            new_df.dropna(subset=["is_valid"], inplace=True)

        return new_df

    def run(self):
        """The main process of the current task."""
        # Import the DataFrame(s)
        if self.dataset_df is None and self.processed_inputs() is None:
            raise ValueError("Either the 'dataset_df' parameter or a requirement must be provided.")

        new_df = self._get_dataset()
        new_df = self._join_inputs(new_df)

        # Create the output directory
        self.output()["data"].pathlib_path.mkdir(parents=True, exist_ok=True)

        # Get args
        if callable(self.args):
            args = self.args()
        else:
            args = self.args
        if not isinstance(args, (list, tuple)):
            raise TypeError(
                "The 'args' must either be a method returning a list or a tuple, or an actual "
                "list or tuple."
            )

        # Get kwargs
        if callable(self.kwargs):
            kwargs = self.kwargs()
        else:
            kwargs = self.kwargs
        if not isinstance(kwargs, dict):
            raise TypeError(
                "The 'kwargs' must either be a method returning a dict, or an actual dict."
            )

        # Add output columns
        if self.output_columns:
            self._report_cols.extend(
                [col for col in self.output_columns if col not in self._report_cols]
            )

        # Format the DataFrame
        df = ValidationResultSet(new_df, output_columns=self.output_columns)

        # Copy current index
        index = df.index.copy()

        # Apply pre process
        self.pre_process(df, args, kwargs)

        # Apply the validation function to the DataFrame
        res = self._process(df, *args, **kwargs)

        # Apply post process
        self.post_process(res, args, kwargs)

        # Count valid items
        self.nb_valid = res["is_valid"].sum()
        self.nb_total = len(res.index)
        self.results = res[list(ValidationResultSet.out_cols.keys())]

        # Export the DataFrame to CSV file
        res[self._report_cols].to_csv(
            self.output()["report"].path, index=True, index_label=INDEX_LABEL
        )

        # Check that the index was not modified
        if not res.index.equals(index):
            raise IndexError(
                "The index changed during the process. Please update your validation function or "
                "your pre/post process functions to avoid this behaviour."
            )

    def output(self):
        """The targets of the current task."""
        class_path = Path(self.task_name)
        return {
            "report": ReportTarget(
                class_path / "report.csv",
                prefix=self.result_path,
                create_parent=False,  # Do not create the parent here because of the tagged output
                task_name=self.task_name,
            ),
            "data": DataDirectoryTarget(
                class_path / self.data_dir,
                prefix=self.result_path,
                create_parent=False,  # Do not create the parent here because of the tagged output
                task_name=self.task_name,
            ),
        }

    @property
    def task_name(self):
        """The name of the task."""
        return self.custom_task_name or self.task_family

    @property
    def __specifications__(self):
        """A specific docstring only used in the report.

        If not overridden, the report uses the usual docstring.
        """
        return self.__doc__


class ElementValidationTask(BaseValidationTask):
    """A class to validate each element of a data set without considering the other elements.

    The ``validation_function`` will receive the row and the output_path as first arguments.
    """

    def _process(self, df, *args, **kwargs):
        return apply_to_df(
            df,
            self.validation_function,
            self.output()["data"].pathlib_path,
            *args,
            nb_processes=self.nb_processes,
            redirect_stdout=self.redirect_stdout,
            **kwargs,
        )


class SetValidationTask(BaseValidationTask):
    """A class to validate an entire data set (useful for global properties).

    The ``validation_function`` will receive the DataFrame and the output_path as first arguments.

    .. note::

        The given :class:`pandas.DataFrame` will always have the columns
        ``["is_valid", "ret_code", "comment", "exception"]``. The ``validation_function``
        should at least update the values for the ``is_valid`` column.
    """

    def _process(self, df, *args, **kwargs):
        try:
            # pylint: disable=not-callable
            self.validation_function(df, self.output()["data"].pathlib_path, *args, **kwargs)
        except Exception:  # pylint: disable=broad-except
            exception = "".join(traceback.format_exception(*sys.exc_info()))
            L.warning("Exception for entire set: %s", exception)
            df["exception"] = exception
            df["is_valid"] = False
            df["ret_code"] = 1

        # Check that return codes are consistent with validity values
        if (
            not df.loc[(df["is_valid"]) & (df["ret_code"] == 1)].empty
            or not df.loc[(~df["is_valid"]) & (df["ret_code"] == 0)].empty
        ):
            raise ValueError("The 'ret_code' values are not consistent with the 'is_valid' values.")
        if not df.loc[(df["comment"].isnull()) & (~df["ret_code"].isin([0, 1]))].empty:
            warnings.warn("A comment should be set when the 'ret_code' is greater than 1.")

        # Fix missing values for return codes according to validity values
        df.loc[(df["is_valid"]) & (df["ret_code"].isnull()), "ret_code"] = 0
        df.loc[(~df["is_valid"]) & (df["ret_code"].isnull()), "ret_code"] = 1

        # Fix missing comment and exception values
        df.loc[df["comment"].isnull(), "comment"] = ""
        df.loc[df["exception"].isnull(), "exception"] = ""

        return df


class ValidationWorkflow(SetValidationTask):
    """Class to define and process a validation workflow."""

    report_path = OptionalStrParameter(
        default=None, description=":str: Path to the workflow report."
    )
    generate_report = BoolParameter(
        default=True, description=":bool: Trigger the report generation."
    )
    report_type = ChoiceParameter(
        default="pdf",
        description=":str: Type of report ('pdf': basic, 'latexpdf': beautiful).",
        choices=["pdf", "latexpdf"],
    )
    report_config = DictParameter(
        default=None,
        description=":dict: The configuration used by Sphinx to build the report.",
    )
    specifications_only = BoolParameter(
        default=False,
        description=":bool: Only outputs the dataset specification document.",
    )

    gather_inputs = True

    def __init__(self, *args, **kwargs):
        event_handler = super().event_handler

        # pylint: disable=unused-variable
        @event_handler(luigi.Event.SUCCESS)
        def spec_report(current_task):
            """Hook to create a specification report."""
            L.debug("Generating report of %s", current_task)
            if current_task.generate_report:
                try:
                    make_report(current_task, config=self.report_config)
                # pylint: disable=broad-except
                except Exception as e:  # pragma: no cover
                    L.error(
                        "The report could not be generated because of the following exception: %s",
                        e,
                    )

        super().__init__(*args, **kwargs)

        if self.specifications_only:
            if not self.report_path:
                self.report_path = f"{self.task_name}_specifications.pdf"
            make_report(self, use_data=False, config=self.report_config)

            self.complete = lambda: True

    @staticmethod
    def validation_function(*args, **kwargs):
        """The validation function to apply to the current data set .

        This method should usually do nothing for :class:`ValidationWorkflow` as this class is only
        supposed to gather validation steps.
        """


def _skippable_element_validation_function(validation_function, skip, *args, **kwargs):
    """Skipping wrapper for an element validation function."""
    if skip:
        return ValidationResult(is_valid=True, comment=SKIP_COMMENT)
    return validation_function(*args, **kwargs)


def _skippable_set_validation_function(validation_function, skip, *args, **kwargs):
    """Skipping wrapper for a set validation function."""
    df = kwargs.get("df", args[0])
    if skip:
        df.loc[df["is_valid"], "comment"] = SKIP_COMMENT
    else:
        validation_function(*args, **kwargs)


def SkippableMixin(default_value=False):
    """Create a mixin class to add a ``skip`` parameter.

    This mixin must be applied to a :class:`data_validation_framework.ElementValidationTask`.
    It will create a ``skip`` parameter and wrap the validation function to just skip it if the
    ``skip`` argument is set to ``True``. If skipped, it will keep the ``is_valid`` values as is and
    add a specific comment to inform the user.

    Args:
        default_value (bool): The default value for the ``skip`` argument.
    """

    class Mixin:
        """A mixin to add a ``skip`` parameter to a :class:`luigi.task`."""

        skip = BoolParameter(default=default_value, description=":bool: Skip the task")

        def __init__(self, *args, **kwargs):

            super().__init__(*args, **kwargs)

            if isinstance(self, ElementValidationTask):
                new_validation_function = partial(
                    _skippable_element_validation_function,
                    self.validation_function,
                    self.skip,
                )
            elif isinstance(self, SetValidationTask) and not isinstance(self, ValidationWorkflow):
                new_validation_function = partial(
                    _skippable_set_validation_function,
                    self.validation_function,
                    self.skip,
                )
            else:
                raise TypeError(
                    "The SkippableMixin can only be associated with children of "
                    "ElementValidationTask or SetValidationTask"
                )
            self._skippable_validation_function = self.validation_function
            self.validation_function = new_validation_function

    return Mixin



"""Test the data_validation_framework.task module."""
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=redefined-outer-name
# pylint: disable=too-many-lines
# pylint: disable=unused-argument
import json
import logging
import re
import sys
import time
from pathlib import Path
from shutil import which

import luigi
import luigi_tools
import numpy as np
import pandas as pd
import pause
import pytest
from diff_pdf_visually import pdfdiff
from luigi.parameter import OptionalStrParameter

from data_validation_framework import report
from data_validation_framework import result
from data_validation_framework import target
from data_validation_framework import task

from . import check_files_exist

SKIP_IF_NO_LATEXMK = not which("latexmk")
REASON_NO_LATEXMK = "The command latexmk is not available."


@pytest.fixture
def dataset_df_path(tmpdir):
    """Create a small CSV dataset and return its path."""
    dataset_df_path = tmpdir / "dataset.csv"
    base_dataset_df = pd.DataFrame({"a": [1, 2], "b": [3, 4]})
    base_dataset_df.to_csv(dataset_df_path)

    return str(dataset_df_path)


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
class TestTagResultOutputMixin:
    """Test the data_validation_framework.task.TagResultOutputMixin class."""

    def test_default(self, tmpdir):
        """Test the simple case."""

        class TestTask(task.TagResultOutputMixin, luigi.Task):

            with_conflict = luigi.BoolParameter(default=False)

            def run(self):
                if not self.tag_output:
                    assert self.output()["without_prefix"].path == str(tmpdir / "file.test")
                    assert self.output()["with_prefix"].path == str(tmpdir / "out" / "file.test")
                elif not self.with_conflict:
                    assert re.match(
                        f"{tmpdir}/out" + r"_\d{8}-\d{2}h\d{2}m\d{2}s/file.test",
                        self.output()["with_prefix"].path,
                    )
                else:
                    assert re.match(
                        f"{tmpdir}/out" + r"_\d{8}-\d{2}h\d{2}m\d{2}s_\d+/file.test",
                        self.output()["with_prefix"].path,
                    )

            def output(self):
                return {
                    "without_prefix": target.TaggedOutputLocalTarget("file.test"),
                    "with_prefix": target.TaggedOutputLocalTarget(
                        "file.test", prefix=self.result_path
                    ),
                }

        root = tmpdir / "out"
        assert luigi.build([TestTask(result_path=str(root))], local_scheduler=True)

        # Test tag name conflicts
        pause.until(int(time.time()) + 1.01)

        t1 = TestTask(result_path=str(root), tag_output=True, with_conflict=False)
        assert luigi.build([t1], local_scheduler=True)

        t2 = TestTask(result_path=str(root), tag_output=True, with_conflict=True)
        assert luigi.build([t2], local_scheduler=True)

    def test_rerun_interaction_left(self, tmpdir):
        """Test the data_validation_framework.task.TagResultOutputMixin class."""
        # pylint: disable=protected-access
        # pylint: disable=too-many-statements
        executed = []

        class TestTaskLeft(luigi_tools.task.RerunMixin, task.TagResultOutputMixin, luigi.Task):

            with_conflict = luigi.BoolParameter(default=False)
            seed = luigi.IntParameter(default=0, significant=False)
            root = None

            def run(self):
                executed.append(
                    (
                        self.result_path,
                        self.output().path,
                        self.with_conflict,
                        self.rerun,
                        self.seed,
                    )
                )
                if not self.tag_output:
                    assert self.output().path == str(self.root / "file.test")
                elif not self.with_conflict:
                    assert re.match(
                        str(self.root) + r"_\d{8}-\d{2}h\d{2}m\d{2}s/file.test", self.output().path
                    )
                else:
                    assert re.match(
                        str(self.root) + r"_\d{8}-\d{2}h\d{2}m\d{2}s_\d+/file.test",
                        self.output().path,
                    )
                with self.output().pathlib_path.open("w", encoding="utf-8") as f:
                    f.writelines(
                        [
                            f"{self.root}\n",
                            f"{self.result_path}\n",
                            f"{self.seed}\n",
                            f"{self.tag_output}\n",
                            f"{self.with_conflict}\n",
                            f"{self.rerun}",
                        ]
                    )

            def output(self):
                return target.TaggedOutputLocalTarget("file.test", prefix=self.result_path)

        expected_files = []

        # ########################################################################### #
        # #################### TEST WITH RERUN MIXIN ON THE LEFT #################### #
        # ########################################################################### #

        # Testing with no tag to create base directory
        TestTaskLeft.root = tmpdir / "out_left"
        target.TaggedOutputLocalTarget._already_changed = False

        t0 = TestTaskLeft(result_path=TestTaskLeft.root, tag_output=False, with_conflict=False)
        assert luigi.build([t0], local_scheduler=True)

        expected_files.extend(
            [
                "out_left",
                "out_left/file.test",
            ]
        )
        check_files_exist(tmpdir, expected_files)

        # Testing with tag, no conflict and rerun
        TestTaskLeft.root = tmpdir / "out_left"
        target.TaggedOutputLocalTarget._already_changed = False

        # Wait for the beginning of the next second
        pause.until(int(time.time()) + 1.01)

        # There is no conflict here because the tasks are defined before any target is created
        assert luigi.build(
            [
                TestTaskLeft(
                    result_path=TestTaskLeft.root, seed=0, tag_output=True, with_conflict=False
                ),
                TestTaskLeft(
                    result_path=TestTaskLeft.root,
                    seed=1,
                    tag_output=True,
                    with_conflict=False,
                    rerun=True,
                ),
            ],
            local_scheduler=True,
        )
        assert len(executed) == 2
        executed.clear()

        expected_files.extend(
            [
                r"out_left_\d{8}-\d{2}h\d{2}m\d{2}s",
                r"out_left_\d{8}-\d{2}h\d{2}m\d{2}s/file.test",
            ]
        )
        check_files_exist(tmpdir, expected_files)

        with Path(
            [
                i
                for i in sorted(Path(tmpdir).rglob("*"))
                if re.match(
                    r"out_left_\d{8}-\d{2}h\d{2}m\d{2}s/file.test", str(i.relative_to(tmpdir))
                )
            ][0]
        ).open("r", encoding="utf-8") as f:
            data = f.read().split("\n")
        assert data[0] == str(tmpdir / "out_left")
        assert re.match(str(tmpdir / r"out_left_\d{8}-\d{2}h\d{2}m\d{2}s"), data[1])
        assert data[2:] == ["1", "True", "False", "True"]

        # Testing with tag, no conflict and rerun
        TestTaskLeft.root = tmpdir / "out_left"
        target.TaggedOutputLocalTarget._already_changed = False

        # Wait for the beginning of the next second
        pause.until(int(time.time()) + 1.01)

        # There is conflict here because the last task is defined after the Rerun mixin which
        # created a new directory
        assert luigi.build(
            [
                TestTaskLeft(
                    result_path=TestTaskLeft.root, seed=0, tag_output=True, with_conflict=True
                ),
                TestTaskLeft(
                    result_path=TestTaskLeft.root,
                    seed=1,
                    tag_output=True,
                    with_conflict=True,
                    rerun=True,
                ),
                TestTaskLeft(
                    result_path=TestTaskLeft.root, seed=2, tag_output=True, with_conflict=True
                ),
            ],
            local_scheduler=True,
        )
        assert len(executed) == 1  # Note that only the last task is executed here
        executed.clear()

        expected_files.extend(
            [
                r"out_left_\d{8}-\d{2}h\d{2}m\d{2}s",
                r"out_left_\d{8}-\d{2}h\d{2}m\d{2}s_\d+",
                r"out_left_\d{8}-\d{2}h\d{2}m\d{2}s_\d+/file.test",
            ]
        )
        check_files_exist(
            tmpdir,
            expected_files,
        )

        with Path(
            [
                i
                for i in sorted(Path(tmpdir).rglob("*"))
                if re.match(
                    r"out_left_\d{8}-\d{2}h\d{2}m\d{2}s_\d+/file.test", str(i.relative_to(tmpdir))
                )
            ][0]
        ).open("r", encoding="utf-8") as f:
            data = f.read().split("\n")
        assert data[0] == str(tmpdir / "out_left")
        assert re.match(str(tmpdir / r"out_left_\d{8}-\d{2}h\d{2}m\d{2}s_\d+"), data[1])
        assert data[2:] == ["2", "True", "True", "False"]

        # Testing with rerun and conflict
        TestTaskLeft.root = tmpdir / "out_left_with_rerun"
        target.TaggedOutputLocalTarget._already_changed = False

        # Wait for the beginning of the next second
        pause.until(int(time.time()) + 1.01)

        # The t1 task has no conflict
        t1 = TestTaskLeft(result_path=TestTaskLeft.root, tag_output=True, with_conflict=False)
        assert t1.output().path.endswith("s/file.test")

        # Test warning is raised when TagResultOutputMixin is used with RerunMixin
        with pytest.warns(
            UserWarning,
            match=(
                "Using 'rerun' with conflicting tag output results in creating a new tag or "
                r"removing the untagged result directory \(depending on the inheritance order\)."
            ),
        ):
            t2 = TestTaskLeft(
                result_path=TestTaskLeft.root, tag_output=True, with_conflict=True, rerun=True
            )

        assert t2.output().path.endswith("s_1/file.test")

        # Build without rerun
        assert luigi.build([t1], local_scheduler=True)
        assert len(executed) == 1
        executed.clear()

        expected_files.extend(
            [
                r"out_left_with_rerun_\d{8}-\d{2}h\d{2}m\d{2}s",
                r"out_left_with_rerun_\d{8}-\d{2}h\d{2}m\d{2}s/file.test",
                r"out_left_with_rerun_\d{8}-\d{2}h\d{2}m\d{2}s_\d+",
            ]
        )
        check_files_exist(
            tmpdir,
            expected_files,
        )

        with Path(
            [
                i
                for i in sorted(Path(tmpdir).rglob("*"))
                if re.match(
                    r"out_left_with_rerun_\d{8}-\d{2}h\d{2}m\d{2}s/file.test",
                    str(i.relative_to(tmpdir)),
                )
            ][0]
        ).open("r", encoding="utf-8") as f:
            data = f.read().split("\n")
        assert data[0] == str(tmpdir / "out_left_with_rerun")
        assert re.match(str(tmpdir / r"out_left_with_rerun_\d{8}-\d{2}h\d{2}m\d{2}s"), data[1])
        assert data[2:] == ["0", "True", "False", "False"]

        # Build with rerun
        assert luigi.build([t2], local_scheduler=True)
        assert len(executed) == 1
        executed.clear()

        expected_files.extend(
            [
                r"out_left_with_rerun_\d{8}-\d{2}h\d{2}m\d{2}s_\d+/file.test",
            ]
        )

        check_files_exist(
            tmpdir,
            expected_files,
        )

        with Path(
            [
                i
                for i in sorted(Path(tmpdir).rglob("*"))
                if re.match(
                    r"out_left_with_rerun_\d{8}-\d{2}h\d{2}m\d{2}s_\d+/file.test",
                    str(i.relative_to(tmpdir)),
                )
            ][0]
        ).open("r", encoding="utf-8") as f:
            data = f.read().split("\n")
        assert data[0] == str(tmpdir / "out_left_with_rerun")
        assert re.match(str(tmpdir / r"out_left_with_rerun_\d{8}-\d{2}h\d{2}m\d{2}s_\d+"), data[1])
        assert data[2:] == ["0", "True", "True", "True"]

    def test_rerun_interaction_right(self, tmpdir):
        """Test the data_validation_framework.task.TagResultOutputMixin class."""
        # pylint: disable=protected-access
        # pylint: disable=too-many-statements
        executed = []

        class TestTaskRight(task.TagResultOutputMixin, luigi_tools.task.RerunMixin, luigi.Task):

            with_conflict = luigi.BoolParameter(default=False)
            seed = luigi.IntParameter(default=0, significant=False)
            root = None

            def run(self):
                executed.append(
                    (
                        self.result_path,
                        self.output().path,
                        self.with_conflict,
                        self.rerun,
                        self.seed,
                    )
                )
                if not self.tag_output:
                    assert self.output().path == str(self.root / "file.test")
                elif self.with_conflict and self.rerun:
                    assert re.match(
                        str(self.root) + r"_\d{8}-\d{2}h\d{2}m\d{2}s_\d+/file.test",
                        self.output().path,
                    )
                else:
                    assert re.match(
                        str(self.root) + r"_\d{8}-\d{2}h\d{2}m\d{2}s/file.test",
                        self.output().path,
                    )
                with self.output().pathlib_path.open("w", encoding="utf-8") as f:
                    f.writelines(
                        [
                            f"{self.root}\n",
                            f"{self.result_path}\n",
                            f"{self.seed}\n",
                            f"{self.tag_output}\n",
                            f"{self.with_conflict}\n",
                            f"{self.rerun}",
                        ]
                    )

            def output(self):
                return target.TaggedOutputLocalTarget("file.test", prefix=self.result_path)

        expected_files = []

        # ########################################################################### #
        # #################### TEST WITH RERUN MIXIN ON THE RIGHT ################### #
        # ########################################################################### #

        # Testing with no tag to create base directory
        TestTaskRight.root = tmpdir / "out_right_no-tag"
        target.TaggedOutputLocalTarget._already_changed = False

        t0 = TestTaskRight(result_path=TestTaskRight.root, tag_output=False, with_conflict=False)
        assert luigi.build([t0], local_scheduler=True)
        assert len(executed) == 1
        executed.clear()

        expected_files.extend(
            [
                "out_right_no-tag",
                "out_right_no-tag/file.test",
            ]
        )
        check_files_exist(
            tmpdir,
            expected_files,
        )

        # Testing with tag, no conflict and rerun
        TestTaskRight.root = tmpdir / "out_right_tag_no-conflict"
        target.TaggedOutputLocalTarget._already_changed = False

        # Wait for the beginning of the next second
        pause.until(int(time.time()) + 1.01)

        # There is no actual conflict here because the tasks are defined before any target is
        # created but the Rerun mixin removes the targets BEFORE the tag is created and creates
        # an empty directory with no tag.
        assert luigi.build(
            [
                TestTaskRight(
                    result_path=TestTaskRight.root, seed=0, tag_output=True, with_conflict=False
                ),
                TestTaskRight(
                    result_path=TestTaskRight.root,
                    seed=1,
                    tag_output=True,
                    with_conflict=False,
                    rerun=True,
                ),
            ],
            local_scheduler=True,
        )
        assert len(executed) == 1  # Note that the first task is not executed here
        executed.clear()

        # expected_files.pop()  # Expect to not find the target of the task without tag
        expected_files.extend(
            [
                r"out_right_tag_no-conflict",
                r"out_right_tag_no-conflict_\d{8}-\d{2}h\d{2}m\d{2}s",
                r"out_right_tag_no-conflict_\d{8}-\d{2}h\d{2}m\d{2}s/file.test",
            ]
        )
        check_files_exist(
            tmpdir,
            expected_files,
        )

        with Path(
            [
                i
                for i in sorted(Path(tmpdir).rglob("*"))
                if re.match(
                    r"out_right_tag_no-conflict_\d{8}-\d{2}h\d{2}m\d{2}s/file.test",
                    str(i.relative_to(tmpdir)),
                )
            ][0]
        ).open("r", encoding="utf-8") as f:
            data = f.read().split("\n")
        assert data[0] == str(tmpdir / "out_right_tag_no-conflict")
        assert re.match(
            str(tmpdir / r"out_right_tag_no-conflict_\d{8}-\d{2}h\d{2}m\d{2}s"), data[1]
        )
        assert data[2:] == ["1", "True", "False", "True"]

        # Testing with tag, no conflict and rerun
        TestTaskRight.root = tmpdir / "out_right_tag_with-conflict_rerun_3-tasks"
        target.TaggedOutputLocalTarget._already_changed = False

        # Wait for the beginning of the next second
        pause.until(int(time.time()) + 1.01)

        # There is conflict here because the Rerun mixin is executed BEFORE the new tag is created
        # (and the Rerun mixin created the parent directory of the targets)
        assert luigi.build(
            [
                TestTaskRight(
                    result_path=TestTaskRight.root, seed=0, tag_output=True, with_conflict=True
                ),
                TestTaskRight(
                    result_path=TestTaskRight.root,
                    seed=1,
                    tag_output=True,
                    with_conflict=True,
                    rerun=True,
                ),
                TestTaskRight(
                    result_path=TestTaskRight.root, seed=2, tag_output=True, with_conflict=True
                ),
            ],
            local_scheduler=True,
        )
        assert len(executed) == 1  # Note that only the last task is executed here
        executed.clear()

        expected_files.extend(
            [
                r"out_right_tag_with-conflict_rerun_3-tasks",
                r"out_right_tag_with-conflict_rerun_3-tasks_\d{8}-\d{2}h\d{2}m\d{2}s",
                r"out_right_tag_with-conflict_rerun_3-tasks_\d{8}-\d{2}h\d{2}m\d{2}s/file.test",
            ]
        )
        check_files_exist(
            tmpdir,
            expected_files,
        )

        with Path(
            [
                i
                for i in sorted(Path(tmpdir).rglob("*"))
                if re.match(
                    r"out_right_tag_with-conflict_rerun_3-tasks_\d{8}-\d{2}h\d{2}m\d{2}s/file.test",
                    str(i.relative_to(tmpdir)),
                )
            ][0]
        ).open("r", encoding="utf-8") as f:
            data = f.read().split("\n")
        assert data[0] == str(tmpdir / "out_right_tag_with-conflict_rerun_3-tasks")
        assert re.match(
            str(tmpdir / r"out_right_tag_with-conflict_rerun_3-tasks_\d{8}-\d{2}h\d{2}m\d{2}s"),
            data[1],
        )
        # Note that this time only the task with seed == 2 is executed!
        assert data[2:] == ["2", "True", "True", "False"]

        # Testing with rerun and conflict
        TestTaskRight.root = tmpdir / "out_right_tag_with-conflict_with-rerun"
        target.TaggedOutputLocalTarget._already_changed = False

        # Wait for the beginning of the next second
        pause.until(int(time.time()) + 1.01)

        # The t1 task has no conflict
        # Note that we have to specify the prefix in the target to avoid conflict. If we do not,
        # the default prefix is updated when the task t2 is defined, so when t1 is built the prefix
        # has been changed and thus the assert fails.
        t1 = TestTaskRight(result_path=TestTaskRight.root, tag_output=True, with_conflict=True)
        assert t1.output().path.endswith("s/file.test")

        # Test warning is raised when TagResultOutputMixin is used with RerunMixin
        with pytest.warns(
            UserWarning,
            match=(
                "Using 'rerun' with conflicting tag output results in creating a new tag or "
                r"removing the untagged result directory \(depending on the inheritance order\)."
            ),
        ):
            t2 = TestTaskRight(
                result_path=TestTaskRight.root, tag_output=True, with_conflict=True, rerun=True
            )

        assert t2.output().path.endswith("s_1/file.test")

        # Build without rerun
        assert luigi.build([t1], local_scheduler=True)
        assert len(executed) == 1
        executed.clear()

        expected_files.extend(
            [
                r"out_right_tag_with-conflict_with-rerun",
                r"out_right_tag_with-conflict_with-rerun_\d{8}-\d{2}h\d{2}m\d{2}s",
                r"out_right_tag_with-conflict_with-rerun_\d{8}-\d{2}h\d{2}m\d{2}s/file.test",
                r"out_right_tag_with-conflict_with-rerun_\d{8}-\d{2}h\d{2}m\d{2}s_\d+",
            ]
        )
        check_files_exist(
            tmpdir,
            expected_files,
        )

        with Path(
            [
                i
                for i in sorted(Path(tmpdir).rglob("*"))
                if re.match(
                    r"out_right_tag_with-conflict_with-rerun_\d{8}-\d{2}h\d{2}m\d{2}s/file.test",
                    str(i.relative_to(tmpdir)),
                )
            ][0]
        ).open("r", encoding="utf-8") as f:
            data = f.read().split("\n")
        assert data[0] == str(tmpdir / "out_right_tag_with-conflict_with-rerun")
        assert re.match(
            str(tmpdir / r"out_right_tag_with-conflict_with-rerun_\d{8}-\d{2}h\d{2}m\d{2}s"),
            data[1],
        )
        assert data[2:] == ["0", "True", "True", "False"]

        # Build with rerun
        assert luigi.build([t2], local_scheduler=True)
        assert len(executed) == 1
        executed.clear()

        expected_files.extend(
            [
                r"out_right_tag_with-conflict_with-rerun_\d{8}-\d{2}h\d{2}m\d{2}s_\d+/file.test",
            ]
        )
        check_files_exist(
            tmpdir,
            expected_files,
        )

        with Path(
            [
                i
                for i in sorted(Path(tmpdir).rglob("*"))
                if re.match(
                    (
                        "out_right_tag_with-conflict_with-rerun_"
                        r"\d{8}-\d{2}h\d{2}m\d{2}s_\d+/file.test"
                    ),
                    str(i.relative_to(tmpdir)),
                )
            ][0]
        ).open("r", encoding="utf-8") as f:
            data = f.read().split("\n")
        assert data[0] == str(tmpdir / "out_right_tag_with-conflict_with-rerun")
        assert re.match(
            str(tmpdir / r"out_right_tag_with-conflict_with-rerun_\d{8}-\d{2}h\d{2}m\d{2}s"),
            data[1],
        )
        assert data[2:] == ["0", "True", "True", "True"]

    def test_propagation(self, tmpdir, dataset_df_path):
        """Test the propagation of result_path."""

        class TestTaskA(task.SetValidationTask):
            """A test validation task."""

            @staticmethod
            def validation_function(df, output_path, *args, **kwargs):
                (output_path / "fileA.test").touch()

        class TestTaskB(task.SetValidationTask):
            """A test validation task."""

            @staticmethod
            def validation_function(df, output_path, *args, **kwargs):
                (output_path / "fileB.test").touch()

        class TestWorkflow(task.ValidationWorkflow):
            """The global validation workflow."""

            with_conflict = luigi.BoolParameter(default=False)

            def inputs(self):
                return {
                    TestTaskA: {},
                    TestTaskB: {},
                }

            def kwargs(self):
                return {
                    "tag_output": self.tag_output,
                    "with_conflict": self.with_conflict,
                }

            @staticmethod
            def validation_function(df, output_path, *args, tag_output, with_conflict, **kwargs):
                res_files = sorted(Path(output_path.parent.parent).rglob("*"))
                if not tag_output:
                    suffix = ""
                elif not with_conflict:
                    suffix = r"_\d{8}-\d{2}h\d{2}m\d{2}s"
                else:
                    suffix = r"_\d{8}-\d{2}h\d{2}m\d{2}s_\d+"
                expected_patterns = [
                    f"{tmpdir}/out" + suffix + "/TestTaskA",
                    f"{tmpdir}/out" + suffix + "/TestTaskA/data",
                    f"{tmpdir}/out" + suffix + "/TestTaskA/data/fileA.test",
                    f"{tmpdir}/out" + suffix + "/TestTaskA/report.csv",
                    f"{tmpdir}/out" + suffix + "/TestTaskB",
                    f"{tmpdir}/out" + suffix + "/TestTaskB/data",
                    f"{tmpdir}/out" + suffix + "/TestTaskB/data/fileB.test",
                    f"{tmpdir}/out" + suffix + "/TestTaskB/report.csv",
                    f"{tmpdir}/out" + suffix + "/TestWorkflow",
                    f"{tmpdir}/out" + suffix + "/TestWorkflow/data",
                ]
                assert len(res_files) == len(expected_patterns)
                for path, pattern in zip(res_files, expected_patterns):
                    assert re.match(str(pattern), str(path))

        root = tmpdir / "out"

        workflow = TestWorkflow(
            dataset_df=dataset_df_path,
            result_path=str(root),
            tag_output=False,
            with_conflict=False,
        )
        assert luigi.build([workflow], local_scheduler=True)

        report_df = pd.read_csv(workflow.result_path / "TestWorkflow" / "report.csv")
        assert report_df["is_valid"].all()

        # Test tag name conflicts
        pause.until(int(time.time()) + 1.01)

        workflow_1 = TestWorkflow(
            dataset_df=dataset_df_path,
            result_path=str(root),
            tag_output=True,
            with_conflict=False,
        )
        assert re.match(
            r"out_\d{8}-\d{2}h\d{2}m\d{2}s/TestWorkflow/report.csv",
            str(workflow_1.output()["report"].pathlib_path.relative_to(tmpdir)),
        )
        # Create the result directory of workflow_1 so that it appears to have already run for
        # workflow_2
        workflow_1.output()["data"].pathlib_path.mkdir(parents=True, exist_ok=True)

        workflow_2 = TestWorkflow(
            dataset_df=dataset_df_path,
            result_path=str(root),
            tag_output=True,
            with_conflict=True,
        )
        assert re.match(
            r"out_\d{8}-\d{2}h\d{2}m\d{2}s_\d+/TestWorkflow/report.csv",
            str(workflow_2.output()["report"].pathlib_path.relative_to(tmpdir)),
        )

        assert luigi.build([workflow_1, workflow_2], local_scheduler=True)

        report_df_1 = pd.read_csv(workflow_1.result_path / "TestWorkflow" / "report.csv")
        assert report_df_1["is_valid"].all()
        report_df_2 = pd.read_csv(workflow_2.result_path / "TestWorkflow" / "report.csv")
        assert report_df_2["is_valid"].all()


class TestSetValidationTask:
    """Test the data_validation_framework.task.SetValidationTask class."""

    @pytest.fixture
    def TestTask(self, tmpdir):
        """Create a test task class."""

        class TestTask(task.SetValidationTask):
            @staticmethod
            def validation_function(df, output_path, *args, **kwargs):
                assert df["is_valid"].all()
                assert (df["ret_code"] == 0).all()
                assert df["comment"].isnull().all()
                assert df["exception"].isnull().all()
                assert df["a"].tolist() == [1, 2]
                assert df["b"].tolist() == [3, 4]
                df["a"] *= 10
                df[["a", "b"]].to_csv(output_path / "test.csv")

        return TestTask

    def test_defaults(self, TestTask, dataset_df_path, tmpdir):
        """Test defaults."""
        assert luigi.build(
            [TestTask(dataset_df=dataset_df_path, result_path=str(tmpdir / "out_defaults"))],
            local_scheduler=True,
        )

        result = pd.read_csv(tmpdir / "out_defaults" / "TestTask" / "data" / "test.csv")
        expected = pd.read_csv(tmpdir / "dataset.csv")
        expected["a"] *= 10
        assert result.equals(expected)

    def test_no_dataset_no_input(self, TestTask, tmpdir):
        """Test with no dataset and no input (should fail)."""

        class FailingTestTask(TestTask):
            pass

        failed_tasks = []
        exceptions = []

        @FailingTestTask.event_handler(luigi.Event.FAILURE)
        def check_exception(failed_task, exception):  # pylint: disable=unused-variable
            failed_tasks.append(str(failed_task))
            exceptions.append(str(exception))

        assert not luigi.build(
            [FailingTestTask(result_path=str(tmpdir / "out_fail"))], local_scheduler=True
        )
        assert failed_tasks == [str(FailingTestTask(result_path=str(tmpdir / "out_fail")))]
        assert exceptions == [
            str(ValueError("Either the 'dataset_df' parameter or a requirement must be provided."))
        ]

    def test_inputs_outputs(self, TestTask, dataset_df_path, tmpdir):
        """Test inputs."""

        class TestTaskWithOutputs(task.SetValidationTask):

            output_columns = {"a": None}

            @staticmethod
            def validation_function(df, output_path, *args, **kwargs):
                assert df["is_valid"].all()
                assert (df["ret_code"] == 0).all()
                assert df["comment"].isnull().all()
                assert df["exception"].isnull().all()
                assert df["a"].tolist() == [1, 2]
                assert df["b"].tolist() == [3, 4]
                df["a"] *= 10
                df[["a", "b"]].to_csv(output_path / "test.csv")

        class TestTaskWithInputs(task.SetValidationTask):
            def inputs(self):
                return {
                    TestTaskWithOutputs: (
                        {
                            "dataset_df": dataset_df_path,
                            "result_path": self.result_path,
                        },
                        {"a": "a_input"},
                    )
                }

            def kwargs(self):
                return {"dataset_df": self.dataset_df}

            @staticmethod
            def validation_function(df, output_path, *args, **kwargs):
                assert df["is_valid"].all()
                assert (df["ret_code"] == 0).all()
                assert df["comment"].isnull().all()
                assert df["exception"].isnull().all()
                assert df["a_input"].tolist() == [10, 20]

                if kwargs["dataset_df"] is not None:
                    assert df["a"].tolist() == [1, 2]
                    assert df["b"].tolist() == [3, 4]

                df["a_input"] *= 10
                df.to_csv(output_path / "test.csv")

        # Test with inputs but no dataset
        assert luigi.build(
            [TestTaskWithInputs(result_path=str(tmpdir / "out_inputs"))], local_scheduler=True
        )

        result = pd.read_csv(tmpdir / "out_inputs" / "TestTaskWithInputs" / "data" / "test.csv")
        expected = pd.read_csv(tmpdir / "dataset.csv")
        expected["a_input"] = expected["a"] * 100
        assert (result["a_input"] == expected["a_input"]).all()

        # Test with inputs and dataset
        assert luigi.build(
            [
                TestTaskWithInputs(
                    dataset_df=dataset_df_path, result_path=str(tmpdir / "out_inputs_and_outputs")
                )
            ],
            local_scheduler=True,
        )

        result = pd.read_csv(
            tmpdir / "out_inputs_and_outputs" / "TestTaskWithInputs" / "data" / "test.csv"
        )
        expected = pd.read_csv(tmpdir / "dataset.csv")
        expected["a_input"] = expected["a"] * 100
        assert result[["a", "b", "a_input"]].equals(expected[["a", "b", "a_input"]])

    def test_missing_columns(self, TestTask, dataset_df_path, tmpdir):
        """Test missing columns in requirements."""

        class TestTaskMissingColumns(task.SetValidationTask):
            def inputs(self):
                return {TestTask: {"a": "a_input"}}

            @staticmethod
            def validation_function(df, output_path, *args, **kwargs):
                assert df["is_valid"].all()
                assert (df["ret_code"] == 0).all()
                assert df["comment"].isnull().all()
                assert df["exception"].isnull().all()
                assert df["a_input"].tolist() == [10, 20]

                df["a_input"] *= 10
                df.to_csv(output_path / "test.csv")

        failed_tasks = []
        exceptions = []

        @TestTaskMissingColumns.event_handler(luigi.Event.FAILURE)
        def check_exception(failed_task, exception):  # pylint: disable=unused-variable
            failed_tasks.append(str(failed_task))
            exceptions.append(str(exception))

        assert not luigi.build(
            [
                TestTaskMissingColumns(
                    dataset_df=dataset_df_path, result_path=str(tmpdir / "out_missing_columns")
                )
            ],
            local_scheduler=True,
        )

        assert failed_tasks == [
            str(
                TestTaskMissingColumns(
                    dataset_df=dataset_df_path, result_path=str(tmpdir / "out_missing_columns")
                )
            )
        ]
        assert exceptions == [
            str(
                KeyError(
                    "The columns ['a'] are missing from the output columns of "
                    "TestTask("
                    f"tag_output=False, result_path={tmpdir / 'out_missing_columns'}, "
                    f"dataset_df={dataset_df_path}, input_index_col=, data_dir=data).",
                )
            )
        ]

    def test_wrong_inputs(self, TestTask, dataset_df_path, tmpdir):
        """Test bad format in inputs."""

        class TestTaskMissingColumns(task.SetValidationTask):
            def inputs(self):
                return {TestTask: "BAD INPUTS"}

            @staticmethod
            def validation_function(df, output_path, *args, **kwargs):
                pass

        with pytest.raises(
            ValueError,
            match=(
                "The input values should either be a dict containing the column mapping or a tuple "
                "with a dict containing the keyword arguments as first element and a dict "
                r"containing the column mapping as second element\."
            ),
        ):
            luigi.build(
                [
                    TestTaskMissingColumns(
                        dataset_df=dataset_df_path, result_path=str(tmpdir / "out_bad_inputs")
                    )
                ],
                local_scheduler=True,
            )

    def test_validation_function_as_method(self, dataset_df_path, tmpdir):
        """Test validation function defined as a method (implicitly casted to staticmethod)."""

        class TestTask(task.SetValidationTask):
            # pylint: disable=no-self-argument
            def validation_function(df, output_path, *args, **kwargs):
                df["a"] *= 10
                df[["a", "b"]].to_csv(output_path / "test.csv")

        # Test defaults
        assert luigi.build(
            [TestTask(dataset_df=dataset_df_path, result_path=str(tmpdir / "out_defaults"))],
            local_scheduler=True,
        )

        result = pd.read_csv(tmpdir / "out_defaults" / "TestTask" / "data" / "test.csv")
        expected = pd.read_csv(tmpdir / "dataset.csv")
        expected["a"] *= 10
        assert result.equals(expected)

    class TestPropagation:
        """Test argument propagation."""

        @pytest.fixture
        def BaseTestTask(self, TestTask):  # noqa: D102
            class BaseTestTask(TestTask):
                def kwargs(self):
                    return {
                        "dataset_df": str(self.dataset_df),
                        "result_path": str(self.result_path),
                    }

                @staticmethod
                def validation_function(df, output_path, *args, **kwargs):
                    with open(output_path / "test.json", "w", encoding="utf-8") as f:
                        json.dump(kwargs, f)

                    TestTask.validation_function(df, output_path, *args, **kwargs)

            return BaseTestTask

        @pytest.fixture
        def TestTaskPassDatasetAndResultPath(self, BaseTestTask):  # noqa: D102
            class TestTaskPassDatasetAndResultPath(task.SetValidationTask):
                def inputs(self):
                    return {BaseTestTask: {}}

                @staticmethod
                def validation_function(df, output_path, *args, **kwargs):
                    assert df["is_valid"].all()
                    assert (df["ret_code"] == 0).all()
                    assert df["comment"].isnull().all()
                    assert df["exception"].isnull().all()
                    assert df["a"].tolist() == [1, 2]

                    df["a"] *= 100
                    df.to_csv(output_path / "test.csv")

            return TestTaskPassDatasetAndResultPath

        @pytest.fixture
        def TestTaskPassDatasetAndResultPathWithKwargs(self, BaseTestTask):  # noqa: D102
            class TestTaskPassDatasetAndResultPath(task.SetValidationTask):
                def inputs(self):
                    return {BaseTestTask: ({"result_path": str(self.result_path / "sub_path")}, {})}

                @staticmethod
                def validation_function(df, output_path, *args, **kwargs):
                    assert df["is_valid"].all()
                    assert (df["ret_code"] == 0).all()
                    assert df["comment"].isnull().all()
                    assert df["exception"].isnull().all()
                    assert df["a"].tolist() == [1, 2]

                    df["a"] *= 100
                    df.to_csv(output_path / "test.csv")

            return TestTaskPassDatasetAndResultPath

        @pytest.fixture
        def TestTaskPassDatasetAndResultPathWithInputParameters(self, BaseTestTask):  # noqa: D102
            class TestTaskPassDatasetAndResultPath(task.SetValidationTask):
                def inputs(self):
                    return {
                        BaseTestTask: task.InputParameters(
                            {}, result_path=str(self.result_path / "sub_path")
                        )
                    }

                @staticmethod
                def validation_function(df, output_path, *args, **kwargs):
                    assert df["is_valid"].all()
                    assert (df["ret_code"] == 0).all()
                    assert df["comment"].isnull().all()
                    assert df["exception"].isnull().all()
                    assert df["a"].tolist() == [1, 2]

                    df["a"] *= 100
                    df.to_csv(output_path / "test.csv")

            return TestTaskPassDatasetAndResultPath

        @staticmethod
        def _check_results(tmpdir, workflow_result_subpath, base_result_subpath=None):

            if base_result_subpath is None:
                base_result_subpath = workflow_result_subpath

            with open(
                tmpdir / base_result_subpath / "BaseTestTask" / "data" / "test.json",
                encoding="utf-8",
            ) as f:
                params = json.load(f)

            assert params == {
                "dataset_df": f"{tmpdir}/dataset.csv",
                "result_path": f"{tmpdir}/{base_result_subpath}",
            }

            result_1 = pd.read_csv(
                tmpdir / base_result_subpath / "BaseTestTask" / "data" / "test.csv"
            )
            result_2 = pd.read_csv(
                tmpdir
                / workflow_result_subpath
                / "TestTaskPassDatasetAndResultPath"
                / "data"
                / "test.csv"
            )
            expected = pd.read_csv(tmpdir / "dataset.csv")
            expected["a"] *= 10
            assert result_1.equals(expected)
            expected["a"] *= 10
            assert result_2[["a", "b"]].equals(expected[["a", "b"]])

        def test_dataset_propagation(
            self, TestTaskPassDatasetAndResultPath, dataset_df_path, tmpdir
        ):
            """Test that the dataset is properly passed to the requirements."""
            assert luigi.build(
                [
                    TestTaskPassDatasetAndResultPath(
                        dataset_df=dataset_df_path, result_path=str(tmpdir / "out_pass_dataset")
                    )
                ],
                local_scheduler=True,
            )

            self._check_results(tmpdir, "out_pass_dataset", "out_pass_dataset")

        def test_dataset_propagation_with_config(
            self, TestTaskPassDatasetAndResultPath, dataset_df_path, tmpdir
        ):
            """Test that the dataset is not propagated when it is already given in the config."""
            with luigi_tools.util.set_luigi_config(
                {
                    "BaseTestTask": {
                        "result_path": str(tmpdir / "specific_out_path"),
                    }
                }
            ):
                assert luigi.build(
                    [
                        TestTaskPassDatasetAndResultPath(
                            dataset_df=dataset_df_path, result_path=str(tmpdir / "out_pass_dataset")
                        )
                    ],
                    local_scheduler=True,
                )

            self._check_results(tmpdir, "out_pass_dataset", "specific_out_path")

        def test_dataset_propagation_with_kwargs(
            self, TestTaskPassDatasetAndResultPathWithKwargs, dataset_df_path, tmpdir
        ):
            """Test that the dataset is properly passed to the requirements."""
            assert luigi.build(
                [
                    TestTaskPassDatasetAndResultPathWithKwargs(
                        dataset_df=dataset_df_path, result_path=str(tmpdir / "out_pass_dataset")
                    )
                ],
                local_scheduler=True,
            )

            self._check_results(tmpdir, "out_pass_dataset", "out_pass_dataset/sub_path")

        def test_dataset_propagation_with_config_and_kwargs(
            self, TestTaskPassDatasetAndResultPathWithKwargs, dataset_df_path, tmpdir
        ):
            """Test that the dataset propagation.

            The dataset should be propagated from the constructor even when a value is already
            given in the config.
            """
            with luigi_tools.util.set_luigi_config(
                {
                    "BaseTestTask": {
                        "result_path": str(tmpdir / "specific_out_path"),
                    }
                }
            ):
                assert luigi.build(
                    [
                        TestTaskPassDatasetAndResultPathWithKwargs(
                            dataset_df=dataset_df_path, result_path=str(tmpdir / "out_pass_dataset")
                        )
                    ],
                    local_scheduler=True,
                )

            self._check_results(tmpdir, "out_pass_dataset", "out_pass_dataset/sub_path")

        def test_dataset_propagation_with_input_parameters(
            self, TestTaskPassDatasetAndResultPathWithInputParameters, dataset_df_path, tmpdir
        ):
            """Test that the dataset is properly passed to the requirements."""
            assert luigi.build(
                [
                    TestTaskPassDatasetAndResultPathWithInputParameters(
                        dataset_df=dataset_df_path, result_path=str(tmpdir / "out_pass_dataset")
                    )
                ],
                local_scheduler=True,
            )

            self._check_results(tmpdir, "out_pass_dataset", "out_pass_dataset/sub_path")

        def test_dataset_propagation_with_config_and_input_parameters(
            self, TestTaskPassDatasetAndResultPathWithInputParameters, dataset_df_path, tmpdir
        ):
            """Test that the dataset propagation.

            The dataset should be propagated from the constructor even when a value is already
            given in the config.
            """
            with luigi_tools.util.set_luigi_config(
                {
                    "BaseTestTask": {
                        "result_path": str(tmpdir / "specific_out_path"),
                    }
                }
            ):
                assert luigi.build(
                    [
                        TestTaskPassDatasetAndResultPathWithInputParameters(
                            dataset_df=dataset_df_path, result_path=str(tmpdir / "out_pass_dataset")
                        )
                    ],
                    local_scheduler=True,
                )

            self._check_results(tmpdir, "out_pass_dataset", "out_pass_dataset/sub_path")

    def test_failing_validation_function(self, TestTask, dataset_df_path, tmpdir):
        """Test with a failing validation function."""

        class FailingTestTask(TestTask):
            @staticmethod
            def validation_function(*args, **kwargs):
                raise RuntimeError("This function always fails")

        assert luigi.build(
            [
                FailingTestTask(
                    dataset_df=dataset_df_path,
                    result_path=str(tmpdir / "out_failing_validation_function"),
                )
            ],
            local_scheduler=True,
        )

        result = pd.read_csv(
            tmpdir / "out_failing_validation_function" / "FailingTestTask" / "report.csv"
        )
        assert (~result["is_valid"]).all()
        assert (result["ret_code"] == 1).all()
        assert (result["comment"].isnull()).all()
        assert (
            result["exception"].str.split("\n").apply(lambda x: x[5])
            == "RuntimeError: This function always fails"
        ).all()

    def test_task_name(self, dataset_df_path, tmpdir):
        """Test with a custom task name."""

        class TestTask(task.SetValidationTask):
            task_name = "test_custom_task_name"

            @staticmethod
            def validation_function(*args, **kwargs):
                pass

        assert luigi.build(
            [
                TestTask(
                    dataset_df=dataset_df_path,
                    result_path=str(tmpdir),
                )
            ],
            local_scheduler=True,
        )
        assert (tmpdir / "test_custom_task_name" / "report.csv").exists()

    def test_duplicated_index(self, tmpdir, TestTask):
        """Test with duplicated index values."""
        dataset_df_path = str(tmpdir / "dataset.csv")
        base_dataset_df = pd.DataFrame({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]}, index=[0, 1, 1, 0])
        base_dataset_df.to_csv(dataset_df_path, index=True, index_label="index_col")

        failed_tasks = []
        exceptions = []

        @TestTask.event_handler(luigi.Event.FAILURE)
        def check_exception(failed_task, exception):  # pylint: disable=unused-variable
            failed_tasks.append(str(failed_task))
            exceptions.append(str(exception))

        failing_task = TestTask(
            dataset_df=dataset_df_path,
            input_index_col="index_col",
            result_path=str(tmpdir / "out_failing_duplicated_index"),
        )
        assert not luigi.build([failing_task], local_scheduler=True)

        assert failed_tasks == [str(failing_task)]
        assert exceptions == [str(IndexError("The following index values are duplicated: [0, 1]"))]

    def test_change_index(self, tmpdir, TestTask):
        """Test that the process fails if the index is changed by the validation function."""
        dataset_df_path = str(tmpdir / "dataset.csv")
        base_dataset_df = pd.DataFrame({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]}, index=[0, 1, 2, 3])
        base_dataset_df.to_csv(dataset_df_path, index=True, index_label="index_col")

        class TestTaskUpdateIndex(TestTask):
            @staticmethod
            def validation_function(df, output_path, *args, **kwargs):
                df.sort_index(ascending=False, inplace=True)

        failed_tasks = []
        exceptions = []

        @TestTaskUpdateIndex.event_handler(luigi.Event.FAILURE)
        def check_exception(failed_task, exception):  # pylint: disable=unused-variable
            failed_tasks.append(str(failed_task))
            exceptions.append(str(exception))

        failing_task = TestTaskUpdateIndex(
            dataset_df=dataset_df_path,
            input_index_col="index_col",
            result_path=str(tmpdir / "out_failing_update_index"),
        )
        assert not luigi.build([failing_task], local_scheduler=True)

        assert failed_tasks == [str(failing_task)]
        assert exceptions == [
            str(
                IndexError(
                    "The index changed during the process. Please update your validation function "
                    "or your pre/post process functions to avoid this behaviour."
                )
            )
        ]

    def test_missing_retcodes(self, tmpdir, dataset_df_path, TestTask):
        """Test invalid retcodes."""

        class TestTaskMissingRetcodes(TestTask):
            mode = OptionalStrParameter(default=None)

            def kwargs(self):
                return {"mode": self.mode}

            @staticmethod
            def validation_function(df, output_path, *args, **kwargs):
                if kwargs["mode"] == "valid":
                    df["is_valid"] = True
                    df["ret_code"] = 1
                elif kwargs["mode"] == "not valid":
                    df["is_valid"] = False
                    df["ret_code"] = 0

        failed_tasks = []
        exceptions = []

        @TestTaskMissingRetcodes.event_handler(luigi.Event.FAILURE)
        def check_exception(failed_task, exception):  # pylint: disable=unused-variable
            failed_tasks.append(str(failed_task))
            exceptions.append(str(exception))

        failing_task_valid = TestTaskMissingRetcodes(
            dataset_df=dataset_df_path,
            result_path=str(tmpdir / "out_missing_retcode"),
            mode="valid",
        )
        failing_task_notvalid = TestTaskMissingRetcodes(
            dataset_df=dataset_df_path,
            result_path=str(tmpdir / "out_missing_retcode"),
            mode="not valid",
        )
        assert not luigi.build([failing_task_valid], local_scheduler=True)
        assert not luigi.build([failing_task_notvalid], local_scheduler=True)

        assert failed_tasks == [str(failing_task_valid), str(failing_task_notvalid)]
        assert (
            exceptions
            == [
                str(
                    task.ValidationError(
                        "The 'ret_code' values are not consistent with the 'is_valid' values."
                    )
                )
            ]
            * 2
        )

    def test_missing_comments(self, tmpdir, dataset_df_path, TestTask):
        """Test missing comments."""

        class TestTaskMissingComments(TestTask):
            mode = OptionalStrParameter(default=None)

            def kwargs(self):
                return {"mode": self.mode}

            @staticmethod
            def validation_function(df, output_path, *args, **kwargs):
                if kwargs["mode"] == "valid":
                    df["is_valid"] = True
                    df["ret_code"] = 2
                elif kwargs["mode"] == "not valid":
                    df["is_valid"] = False
                    df["ret_code"] = 2

        failed_tasks = []
        exceptions = []

        @TestTaskMissingComments.event_handler(luigi.Event.FAILURE)
        def check_exception(failed_task, exception):  # pylint: disable=unused-variable
            failed_tasks.append(str(failed_task))
            exceptions.append(str(exception))

        failing_task_valid = TestTaskMissingComments(
            dataset_df=dataset_df_path,
            result_path=str(tmpdir / "out_missing_comments_valid"),
            mode="valid",
        )
        failing_task_notvalid = TestTaskMissingComments(
            dataset_df=dataset_df_path,
            result_path=str(tmpdir / "out_missing_comments_not_valid"),
            mode="not valid",
        )
        with pytest.warns(
            UserWarning, match="A comment should be set when the 'ret_code' is greater than 1."
        ):
            assert luigi.build([failing_task_valid], local_scheduler=True)
        with pytest.warns(
            UserWarning, match="A comment should be set when the 'ret_code' is greater than 1."
        ):
            assert luigi.build([failing_task_notvalid], local_scheduler=True)

    def test_rename_multiindex(self):
        """Test multiindex renaming."""
        df_1_level = pd.DataFrame({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]}, index=[0, 1, 2, 3])

        task.BaseValidationTask._rename_cols(df_1_level)  # pylint: disable=protected-access
        assert df_1_level.columns.tolist() == ["a", "b"]

        df_2_levels = pd.DataFrame(
            {"a": [1, 2, 3, 4], "b": [5, 6, 7, 8], "Unnamed: c": [9, 10, 11, 12]},
            index=[0, 1, 2, 3],
        )
        df_2_levels.columns = pd.MultiIndex.from_tuples(
            [("level_1", i) for i in df_2_levels.columns.values]
        )
        task.BaseValidationTask._rename_cols(df_2_levels)  # pylint: disable=protected-access
        assert df_2_levels.columns.tolist() == [("level_1", "a"), ("level_1", "b"), ("level_1", "")]

    def test_check_inputs(self, TestTask):
        """Test that inputs are properly checked."""
        assert not task.BaseValidationTask.check_inputs({})
        assert not task.BaseValidationTask.check_inputs(None)

        with pytest.raises(
            ValueError,
            match=(
                r"The destination column of the task TestTask\(tag_output=False, result_path=, "
                r"dataset_df=, input_index_col=, data_dir=data\) can not be one of "
                r"\['is_valid', 'ret_code', 'comment', 'exception'\]\."
            ),
        ):
            task.BaseValidationTask.check_inputs(
                {
                    TestTask(): {"a": "is_valid"},
                }
            )

        class TestConflictingColumns(task.SetValidationTask):
            pass

        with pytest.raises(
            ValueError,
            match=(
                r"The destination column 'a_conflict' of the tasks TestConflictingColumns\("
                r"tag_output=False, result_path=, dataset_df=, input_index_col=, data_dir=data\) "
                r"and TestTask\(tag_output=False, result_path=, dataset_df=, input_index_col=, "
                r"data_dir=data\) are conflicting\."
            ),
        ):
            task.BaseValidationTask.check_inputs(
                {TestTask(): {"a": "a_conflict"}, TestConflictingColumns(): {"a": "a_conflict"}}
            )

    def test_extra_requires(self, tmpdir, dataset_df_path):
        """Test extra_requires feature."""

        class TestTaskA(luigi.Task):
            def run(self):
                assert self.output().path == str(tmpdir / "file.test")
                with open(self.output().path, "w", encoding="utf-8") as f:
                    f.write("result of TestTaskA")

            def output(self):
                return target.OutputLocalTarget(tmpdir / "file.test")

        class TestTaskB(task.SetValidationTask):

            output_columns = {"extra_path": None, "extra_result": None}

            def kwargs(self):
                return {"extra_task_target": self.extra_input().path}

            def extra_requires(self):
                return TestTaskA()

            @staticmethod
            def validation_function(df, output_path, *args, **kwargs):
                df["is_valid"] = True
                df["extra_path"] = kwargs["extra_task_target"]
                with open(kwargs["extra_task_target"], encoding="utf-8") as f:
                    df["extra_result"] = f.read()

        assert luigi.build(
            [TestTaskB(dataset_df=dataset_df_path, result_path=str(tmpdir / "extra_requires"))],
            local_scheduler=True,
        )

        res = pd.read_csv(tmpdir / "extra_requires" / "TestTaskB" / "report.csv")
        assert (res["extra_path"] == str(tmpdir / "file.test")).all()
        assert (res["extra_result"] == "result of TestTaskA").all()

    def test_static_args_kwargs(self, dataset_df_path):
        """Test the args and kwargs feature."""

        class TestTask(task.ElementValidationTask):

            args = [1, "a"]
            kwargs = {"int_value": 1, "str_value": "a"}

            @staticmethod
            def validation_function(df, output_path, *args, **kwargs):
                assert args == [1, "a"]
                assert kwargs == {"int_value": 1, "str_value": "a"}

        assert luigi.build(
            [TestTask(dataset_df=dataset_df_path)],
            local_scheduler=True,
        )

        class TestFailingArgsTask(task.ElementValidationTask):

            args = 1

            @staticmethod
            def validation_function(df, output_path, *args, **kwargs):
                assert args == 1

        failed_tasks = []
        exceptions = []

        @TestFailingArgsTask.event_handler(luigi.Event.FAILURE)
        def check_exception_args(failed_task, exception):  # pylint: disable=unused-variable
            failed_tasks.append(str(failed_task))
            exceptions.append(str(exception))

        assert not luigi.build(
            [TestFailingArgsTask(dataset_df=dataset_df_path)],
            local_scheduler=True,
        )

        assert failed_tasks == [str(TestFailingArgsTask(dataset_df=dataset_df_path))]
        assert exceptions == [
            str(
                TypeError(
                    "The 'args' must either be a method returning a list or a tuple, or an actual "
                    "list or tuple."
                )
            )
        ]

        class TestFailingKwargsTask(task.ElementValidationTask):

            kwargs = 1

            @staticmethod
            def validation_function(df, output_path, *args, **kwargs):
                assert kwargs == 1

        failed_tasks = []
        exceptions = []

        @TestFailingKwargsTask.event_handler(luigi.Event.FAILURE)
        def check_exception_kwargs(failed_task, exception):  # pylint: disable=unused-variable
            failed_tasks.append(str(failed_task))
            exceptions.append(str(exception))

        assert not luigi.build(
            [TestFailingKwargsTask(dataset_df=dataset_df_path)],
            local_scheduler=True,
        )

        assert failed_tasks == [str(TestFailingKwargsTask(dataset_df=dataset_df_path))]
        assert exceptions == [
            str(
                TypeError(
                    "The 'kwargs' must either be a method returning a dict, or an actual dict."
                )
            )
        ]

    def test_different_input_index(self, tmpdir, dataset_df_path, caplog):
        """Test inconsistent indexes between the dataset and the input tasks."""

        class TestTask(task.ElementValidationTask):
            @staticmethod
            def validation_function(df, output_path, *args, **kwargs):
                pass

        class TestDifferentInputIndex(task.ElementValidationTask):
            def inputs(self):
                return {
                    TestTask: (
                        {
                            "dataset_df": dataset_df_path,
                            "input_index_col": self.input_index_col,
                        },
                        {},
                    )
                }

            @staticmethod
            def validation_function(df, output_path, *args, **kwargs):
                pass

        updated_dataset = pd.read_csv(dataset_df_path)
        updated_dataset.index += 1
        name_parts = dataset_df_path.rsplit(".", 1)
        name_parts[0] += "_updated"
        new_df_path = ".".join(name_parts)
        updated_dataset.to_csv(new_df_path)
        caplog.clear()
        caplog.set_level(logging.DEBUG)
        assert luigi.build(
            [
                TestDifferentInputIndex(
                    dataset_df=new_df_path,
                    result_path=str(tmpdir / "different_input_index"),
                    input_index_col=0,
                )
            ],
            local_scheduler=True,
        )
        res = [
            i
            for i in caplog.record_tuples
            if i[0] == "data_validation_framework.task" and i[1] == logging.WARNING
        ]
        assert len(res) == 1
        assert res[0][2] == (
            "The following inconsistent indexes between the dataset and the inputs are "
            "ignored: [2]"
        )

    def test_external_function(self, dataset_df_path):
        """Test with an external function."""

        def external_function(df, output_path, *args, **kwargs):
            assert args == [1, "a"]
            assert kwargs == {"k1": 1, "k2": 2}

        class TestExternalFunctionTask(task.ElementValidationTask):

            args = [1, "a"]
            kwargs = {"k1": 1, "k2": 2}
            validation_function = external_function

        assert luigi.build(
            [TestExternalFunctionTask(dataset_df=dataset_df_path)],
            local_scheduler=True,
        )


class TestElementValidationTask:
    """Test the data_validation_framework.task.ElementValidationTask class."""

    @pytest.fixture
    def TestTask(self, tmpdir):  # noqa: D102
        class TestTask(task.ElementValidationTask):
            @staticmethod
            # pylint: disable=arguments-differ
            def validation_function(row, output_path, *args, **kwargs):
                if row["a"] <= 1:
                    return result.ValidationResult(is_valid=True)
                if row["a"] <= 2:
                    return result.ValidationResult(is_valid=False, comment="bad value")
                raise ValueError(f"Incorrect value {row['a']}")

        return TestTask

    @pytest.fixture
    def dataset_df_path(self, tmpdir):
        """Create a small CSV dataset and return its path."""
        dataset_df_path = tmpdir / "dataset.csv"
        base_dataset_df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
        base_dataset_df.to_csv(dataset_df_path)

        return str(dataset_df_path)

    def test_defaults(self, TestTask, dataset_df_path, tmpdir):
        """Test that the dataset is properly passed to the requirements."""
        assert luigi.build(
            [TestTask(dataset_df=dataset_df_path, result_path=str(tmpdir / "out"))],
            local_scheduler=True,
        )
        result = pd.read_csv(tmpdir / "out" / "TestTask" / "report.csv")
        assert result["is_valid"].tolist() == [True, False, False]
        assert result["ret_code"].tolist() == [0, 1, 1]
        assert result["comment"].tolist() == [np.nan, "bad value", np.nan]
        assert result.loc[[0, 1], "exception"].isnull().all()
        line_shift = sys.version_info.minor >= 3 and sys.version_info.minor >= 11
        assert (
            result.loc[2, "exception"].split("\n")[5 + line_shift]
            == "ValueError: Incorrect value 3"
        )

    @pytest.mark.parametrize("nb_processes", [None, 1, 5])
    def test_nb_processes(self, TestTask, dataset_df_path, tmpdir, nb_processes):
        """Test that the number of processes is properly passed to the requirements."""

        class TestWorkflow(task.ValidationWorkflow):
            def inputs(self):
                return {TestTask: {}}

        assert (
            TestWorkflow(
                dataset_df=dataset_df_path,
                result_path=str(tmpdir / "out"),
                nb_processes=nb_processes,
            )
            .requires()[0]
            .nb_processes
            == nb_processes
        )

    @pytest.mark.parametrize("redirect_stdout", [True, False])
    def test_redirect_stdout(self, TestTask, dataset_df_path, tmpdir, redirect_stdout):
        """Test that the number of processes is properly passed to the requirements."""

        class TestWorkflow(task.ValidationWorkflow):
            def inputs(self):
                return {TestTask: {}}

        assert (
            TestWorkflow(
                dataset_df=dataset_df_path,
                result_path=str(tmpdir / "out"),
                redirect_stdout=redirect_stdout,
            )
            .requires()[0]
            .redirect_stdout
            == redirect_stdout
        )

    @pytest.fixture
    def empty_dataset_df_path(self, tmpdir):
        """Create an empty CSV dataset and return its path."""
        dataset_df_path = tmpdir / "dataset.csv"
        base_dataset_df = pd.DataFrame({"a": [], "b": []})
        base_dataset_df.to_csv(dataset_df_path)

        return str(dataset_df_path)

    def test_defaults_with_empty(self, TestTask, empty_dataset_df_path, tmpdir):
        """Test that the empty dataset is properly processed."""
        assert luigi.build(
            [TestTask(dataset_df=empty_dataset_df_path, result_path=str(tmpdir / "out"))],
            local_scheduler=True,
        )
        result = pd.read_csv(tmpdir / "out" / "TestTask" / "report.csv")
        assert result["is_valid"].tolist() == []
        assert result["ret_code"].tolist() == []
        assert result["comment"].tolist() == []
        assert result["exception"].tolist() == []


class TestValidationWorkflow:
    """Test the data_validation_framework.task.ValidationWorkflow class."""

    @pytest.fixture
    def comment(self):
        """Return a comment."""
        return "This element is not valid"

    @pytest.fixture
    def exception(self):
        """Return an exception message."""
        return (
            "Traceback (most recent call last):"
            '  File "test.py", line 1, in <module>'
            '    raise ValueError("This element is not valid")'
            "ValueError: This element is not valid"
        )

    @pytest.fixture
    def default_report_config_test_date(self):
        """Set a default date in the report config."""
        report._DEFAULT_REPORT_CONFIG["today"] = "TEST DATE"  # pylint: disable=protected-access
        yield None
        report._DEFAULT_REPORT_CONFIG.pop("today")  # pylint: disable=protected-access

    def test_gather(self, tmpdir, dataset_df_path, comment, exception):
        """Test that the results are properly gathered in a workflow."""

        class TestTask(task.SetValidationTask):
            @staticmethod
            def validation_function(df, output_path, *args, **kwargs):
                df.loc[1, "is_valid"] = False
                df.loc[1, "ret_code"] = 1
                df.loc[1, "comment"] = comment
                df.loc[1, "exception"] = exception

        class TestWorkflow(task.ValidationWorkflow):
            def inputs(self):
                return {TestTask: {}}

            @staticmethod
            def validation_function(df, output_path, *args, **kwargs):
                df.to_csv(output_path / "test_gather.csv")

        assert luigi.build(
            [TestWorkflow(dataset_df=dataset_df_path, result_path=str(tmpdir))],
            local_scheduler=True,
        )

        assert (tmpdir / "dataset.csv").exists()
        assert (tmpdir / "TestTask" / "report.csv").exists()
        assert (tmpdir / "TestWorkflow" / "report.csv").exists()

        result = pd.read_csv(tmpdir / "TestWorkflow" / "report.csv")
        expected = pd.DataFrame(
            {
                "__index_label__": [0, 1],
                "is_valid": [True, False],
                "ret_code": [0, 1],
                "('TestTask', 'is_valid')": [True, False],
                "('TestTask', 'ret_code')": [0, 1],
            }
        )
        assert result.columns.tolist() == [
            "__index_label__",
            "is_valid",
            "ret_code",
            "comment",
            "exception",
            "('TestTask', 'is_valid')",
            "('TestTask', 'ret_code')",
            "('TestTask', 'comment')",
            "('TestTask', 'exception')",
        ]
        assert result[  # pylint: disable=unsubscriptable-object
            [
                "__index_label__",
                "is_valid",
                "ret_code",
                "('TestTask', 'is_valid')",
                "('TestTask', 'ret_code')",
            ]
        ].equals(expected)
        assert (
            result.loc[
                0, ["comment", "exception", "('TestTask', 'comment')", "('TestTask', 'exception')"]
            ]
            .isnull()
            .all()
            .all()
        )
        assert (result.loc[1, ["comment", "exception"]].isnull()).all()
        assert result.loc[1, "('TestTask', 'comment')"] == comment
        assert result.loc[1, "('TestTask', 'exception')"] == exception

    def test_no_report(self, tmpdir, dataset_df_path):
        """Test no report generation feature."""

        class TestTask(task.SetValidationTask):
            @staticmethod
            def validation_function(*args, **kwargs):
                pass

        class TestWorkflow(task.ValidationWorkflow):
            generate_report = False

            def inputs(self):
                return {TestTask: {}}

        assert luigi.build(
            [TestWorkflow(dataset_df=dataset_df_path, result_path=str(tmpdir))],
            local_scheduler=True,
        )
        assert (tmpdir / "TestWorkflow" / "report.csv").exists()
        assert not (tmpdir / "report.pdf").exists()

    class TestReport:
        """Test the report generation after workflow run."""

        @pytest.fixture
        def TestTask(self, comment, exception):  # noqa: D102
            class TestTask(task.SetValidationTask):
                """A test validation task."""

                no_exception = luigi.BoolParameter(default=False)
                mode = OptionalStrParameter(default=None)

                def kwargs(self):
                    return {
                        "no_exception": self.no_exception,
                        "mode": self.mode,
                    }

                @staticmethod
                def validation_function(df, output_path, *args, **kwargs):
                    if kwargs["mode"] == "all_succeed":
                        df["is_valid"] = True
                        df["ret_code"] = 0
                    elif kwargs["mode"] == "all_fail":
                        df["is_valid"] = False
                        df["ret_code"] = 1
                    else:
                        df.loc[1, "is_valid"] = False
                        df.loc[1, "ret_code"] = 1
                        if not kwargs["no_exception"]:
                            df.loc[1, "comment"] = comment
                            df.loc[1, "exception"] = exception

            return TestTask

        @pytest.fixture
        def TestTask_Specifications(self):  # noqa: D102
            class TestTask_Specifications(task.SetValidationTask):
                """A test validation task with a specific report doc."""

                __specifications__ = "The specific doc only used in report."

                @staticmethod
                def validation_function(*args, **kwargs):
                    pass

            return TestTask_Specifications

        @pytest.fixture
        def TestWorkflow(self, TestTask, TestTask_Specifications):  # noqa: D102
            class TestWorkflow(task.ValidationWorkflow):
                """The global validation workflow."""

                no_exception = luigi.BoolParameter(default=False)
                mode = OptionalStrParameter(default=None)

                def inputs(self):
                    return {
                        TestTask: ({"no_exception": self.no_exception, "mode": self.mode}, {}),
                        TestTask_Specifications: {},
                    }

            return TestWorkflow

        def test_specifications(self, TestTask, TestTask_Specifications):
            """Test spcifications different from the docstring of the task."""
            assert TestTask(dataset_df="", result_path="").__specifications__ == (
                "A test validation task."
            )
            assert TestTask_Specifications(dataset_df="", result_path="").__specifications__ == (
                "The specific doc only used in report."
            )

        def test_rst2pdf(
            self, tmpdir, dataset_df_path, data_dir, TestWorkflow, default_report_config_test_date
        ):
            """Test PDF generation using rst2pdf."""
            root = tmpdir / "rst2pdf"
            assert luigi.build(
                [TestWorkflow(dataset_df=dataset_df_path, result_path=str(root))],
                local_scheduler=True,
            )
            assert (root / "TestWorkflow" / "report.csv").exists()
            assert (root / "report_TestWorkflow.pdf").exists()
            assert pdfdiff(
                root / "report_TestWorkflow.pdf", data_dir / "test_report" / "report_rst2pdf.pdf"
            )

        @pytest.mark.skipif(SKIP_IF_NO_LATEXMK, reason=REASON_NO_LATEXMK)
        def test_latexpdf(
            self, tmpdir, dataset_df_path, data_dir, TestWorkflow, default_report_config_test_date
        ):
            """Test PDF generation using latexpdf."""
            root = tmpdir / "latexpdf"
            assert luigi.build(
                [
                    TestWorkflow(
                        dataset_df=dataset_df_path, result_path=str(root), report_type="latexpdf"
                    )
                ],
                local_scheduler=True,
            )
            assert (root / "TestWorkflow" / "report.csv").exists()
            assert (root / "report_TestWorkflow.pdf").exists()
            assert pdfdiff(
                root / "report_TestWorkflow.pdf", data_dir / "test_report" / "report_latexpdf.pdf"
            )

        def test_fail_element_no_exception(
            self, tmpdir, dataset_df_path, data_dir, TestWorkflow, default_report_config_test_date
        ):
            """Test PDF generation with exception message and comment."""
            root = tmpdir / "rst2pdf_no_exception"
            assert luigi.build(
                [
                    TestWorkflow(
                        dataset_df=dataset_df_path, result_path=str(root), no_exception=True
                    )
                ],
                local_scheduler=True,
            )
            assert (root / "TestWorkflow" / "report.csv").exists()
            assert (root / "report_TestWorkflow.pdf").exists()
            assert pdfdiff(
                root / "report_TestWorkflow.pdf",
                data_dir / "test_report" / "report_no_exception_rst2pdf.pdf",
            )

        def test_exception_levels(
            self, tmpdir, dataset_df_path, data_dir, TestWorkflow, default_report_config_test_date
        ):
            """Test workflow with several nested exception levels."""

            class TestTask_levels(task.SetValidationTask):
                """A test validation task with a deep level."""

                @staticmethod
                def validation_function(*args, **kwargs):
                    raise ValueError("Bad value")

            class TestWorkflow_lvl4(task.ValidationWorkflow):
                """A nested validation workflow with level 3."""

                def inputs(self):
                    return {
                        TestTask_levels: {},
                    }

            class TestWorkflow_lvl3(task.ValidationWorkflow):
                """A nested validation workflow with level 3."""

                def inputs(self):
                    return {
                        TestWorkflow_lvl4: {},
                    }

            class TestWorkflow_lvl2(task.ValidationWorkflow):
                """A nested validation workflow with level 2."""

                def inputs(self):
                    return {
                        TestWorkflow_lvl3: {},
                    }

            class TestWorkflow_lvl1(task.ValidationWorkflow):
                """A nested validation workflow with level 1."""

                def inputs(self):
                    return {
                        TestWorkflow_lvl2: {},
                    }

            class TestWorkflow_lvl0(task.ValidationWorkflow):
                """The global validation workflow."""

                def inputs(self):
                    return {
                        TestWorkflow_lvl1: {},
                    }

            root = tmpdir / "rst2pdf_levels"
            assert luigi.build(
                [TestWorkflow_lvl0(dataset_df=dataset_df_path, result_path=str(root))],
                local_scheduler=True,
            )
            assert (root / "TestWorkflow_lvl0" / "report.csv").exists()
            assert (root / "report_TestWorkflow_lvl4.pdf").exists()
            assert (root / "report_TestWorkflow_lvl3.pdf").exists()
            assert (root / "report_TestWorkflow_lvl2.pdf").exists()
            assert (root / "report_TestWorkflow_lvl1.pdf").exists()
            assert (root / "report_TestWorkflow_lvl0.pdf").exists()

        def test_report_relative_path(
            self, tmpdir, dataset_df_path, data_dir, TestWorkflow, default_report_config_test_date
        ):
            """Test with a relative report path."""
            root = tmpdir / "relative_path"
            assert luigi.build(
                [
                    TestWorkflow(
                        dataset_df=dataset_df_path,
                        result_path=str(root),
                        report_path="report_test_name.pdf",
                    )
                ],
                local_scheduler=True,
            )
            assert (root / "TestWorkflow" / "report.csv").exists()
            assert (root / "report_test_name.pdf").exists()
            assert pdfdiff(
                root / "report_test_name.pdf", data_dir / "test_report" / "report_rst2pdf.pdf"
            )

        def test_report_absolute_path(
            self, tmpdir, dataset_df_path, data_dir, TestWorkflow, default_report_config_test_date
        ):
            """Test with an absolute report path."""
            root = tmpdir / "absolute_path"
            assert luigi.build(
                [
                    TestWorkflow(
                        dataset_df=dataset_df_path,
                        result_path=str(root),
                        report_path=str(tmpdir / "report_test_absolute.pdf"),
                    )
                ],
                local_scheduler=True,
            )
            assert (root / "TestWorkflow" / "report.csv").exists()
            assert (tmpdir / "report_test_absolute.pdf").exists()
            assert pdfdiff(
                tmpdir / "report_test_absolute.pdf", data_dir / "test_report" / "report_rst2pdf.pdf"
            )

        def test_report_all_succeed(
            self, tmpdir, dataset_df_path, data_dir, TestWorkflow, default_report_config_test_date
        ):
            """Test report generation when all tasks succeed."""
            root = tmpdir / "all_succeed"
            assert luigi.build(
                [
                    TestWorkflow(
                        dataset_df=dataset_df_path,
                        result_path=str(root),
                        report_path=str(tmpdir / "report_test_all_success.pdf"),
                        mode="all_succeed",
                    )
                ],
                local_scheduler=True,
            )
            assert (root / "TestWorkflow" / "report.csv").exists()
            assert (tmpdir / "report_test_all_success.pdf").exists()
            assert pdfdiff(
                tmpdir / "report_test_all_success.pdf",
                data_dir / "test_report" / "report_rst2pdf_all_success.pdf",
            )

        def test_report_all_fail(
            self, tmpdir, dataset_df_path, data_dir, TestWorkflow, default_report_config_test_date
        ):
            """Test report generation when all tasks fail."""
            root = tmpdir / "all_fail"
            assert luigi.build(
                [
                    TestWorkflow(
                        dataset_df=dataset_df_path,
                        result_path=str(root),
                        report_path=str(tmpdir / "report_test_all_fail.pdf"),
                        mode="all_fail",
                    )
                ],
                local_scheduler=True,
            )
            assert (root / "TestWorkflow" / "report.csv").exists()
            assert (tmpdir / "report_test_all_fail.pdf").exists()
            assert pdfdiff(
                tmpdir / "report_test_all_fail.pdf",
                data_dir / "test_report" / "report_rst2pdf_all_fail.pdf",
            )

        def test_report_warnings(
            self, tmpdir, dataset_df_path, data_dir, default_report_config_test_date
        ):
            """Test report generation with warnings."""

            class TestTask_Warning(task.ElementValidationTask):
                """A test validation task which can return warnings."""

                @staticmethod
                # pylint: disable=arguments-differ
                def validation_function(row, output_path, *args, **kwargs):
                    if row["a"] <= 1:
                        return result.ValidationResult(
                            is_valid=True, ret_code=2, comment="a succeeding warning"
                        )
                    if row["a"] <= 2:
                        return result.ValidationResult(
                            is_valid=False, ret_code=3, comment="a failing warning"
                        )
                    raise ValueError(f"Incorrect value {row['a']}")

            class TestWorkflow(task.ValidationWorkflow):
                """The global validation workflow."""

                def inputs(self):
                    return {
                        TestTask_Warning: {},
                    }

            root = tmpdir / "with_warnings"
            assert luigi.build(
                [
                    TestWorkflow(
                        dataset_df=dataset_df_path,
                        result_path=str(root),
                        report_path=str(tmpdir / "report_test_warnings.pdf"),
                    )
                ],
                local_scheduler=True,
            )
            assert (root / "TestWorkflow" / "report.csv").exists()
            assert (tmpdir / "report_test_warnings.pdf").exists()
            assert pdfdiff(
                tmpdir / "report_test_warnings.pdf",
                data_dir / "test_report" / "report_rst2pdf_warnings.pdf",
            )

    class TestReportBeforeRun:
        """Test the report generation before workflow run (generate only the specifications)."""

        @pytest.fixture
        def TestTask(self):  # noqa: D102
            class TestTask(task.SetValidationTask):
                """A test validation task."""

                def run(self):
                    raise RuntimeError("THIS TASK SHOULD NOT BE RUN")

            return TestTask

        @pytest.fixture
        def TestTask_Specifications(self):  # noqa: D102
            class TestTask_Specifications(task.SetValidationTask):
                """A test validation task with a specific report doc."""

                __specifications__ = "The specific doc only used in report."

                def run(self):
                    raise RuntimeError("THIS TASK SHOULD NOT BE RUN")

            return TestTask_Specifications

        @pytest.fixture
        def TestWorkflow(self, TestTask, TestTask_Specifications):  # noqa: D102
            class TestWorkflow(task.ValidationWorkflow):
                """The global validation workflow."""

                def inputs(self):
                    return {
                        TestTask: {},
                        TestTask_Specifications: {},
                    }

            return TestWorkflow

        @pytest.fixture
        def TestNestedWorkflow(self, TestTask, TestWorkflow):  # noqa: D102
            class TestNestedWorkflow(task.ValidationWorkflow):
                """The global validation workflow."""

                def inputs(self):
                    return {
                        TestTask: {},
                        TestWorkflow: {},
                    }

            return TestNestedWorkflow

        def test_rst2pdf(self, tmpdir, dataset_df_path, data_dir, TestWorkflow):
            """Test PDF generation using rst2pdf."""
            root = tmpdir / "rst2pdf"
            assert luigi.build(
                [
                    TestWorkflow(
                        dataset_df=dataset_df_path,
                        result_path=str(root),
                        specifications_only=True,
                    )
                ],
                local_scheduler=True,
            )

            assert (root / "TestWorkflow_specifications.pdf").exists()
            assert pdfdiff(
                root / "TestWorkflow_specifications.pdf",
                data_dir / "test_report_before_run" / "report_rst2pdf.pdf",
                threshold=25,
            )

        def test_rst2pdf_report_path(self, tmpdir, dataset_df_path, data_dir, TestWorkflow):
            """Test PDF generation using rst2pdf and specifying a report path."""
            root = tmpdir / "rst2pdf"
            assert luigi.build(
                [
                    TestWorkflow(
                        dataset_df=dataset_df_path,
                        result_path=str(root),
                        report_path="test_custom_document_name.pdf",
                        specifications_only=True,
                    )
                ],
                local_scheduler=True,
            )

            assert (root / "test_custom_document_name.pdf").exists()
            assert pdfdiff(
                root / "test_custom_document_name.pdf",
                data_dir / "test_report_before_run" / "report_rst2pdf.pdf",
                threshold=25,
            )

        @pytest.mark.skipif(SKIP_IF_NO_LATEXMK, reason=REASON_NO_LATEXMK)
        def test_latexpdf(self, tmpdir, dataset_df_path, data_dir, TestWorkflow):
            """Test PDF generation using latexpdf."""
            root = tmpdir / "latexpdf"
            assert luigi.build(
                [
                    TestWorkflow(
                        dataset_df=dataset_df_path,
                        result_path=str(root),
                        report_type="latexpdf",
                        specifications_only=True,
                    )
                ],
                local_scheduler=True,
            )

            assert (root / "TestWorkflow_specifications.pdf").exists()
            assert pdfdiff(
                root / "TestWorkflow_specifications.pdf",
                data_dir / "test_report_before_run" / "report_latexpdf.pdf",
                threshold=15,
            )

        @pytest.fixture
        def report_config(self):
            """Create a dummy report config."""
            return {
                "project": "Test title",
                "version": "999",
                "author": "Test author",
                "today": "FIXED DATE FOR TESTS",
            }

        def test_rst2pdf_with_config(
            self, tmpdir, dataset_df_path, data_dir, TestWorkflow, report_config
        ):
            """Test PDF generation using rst2pdf with a specific config."""
            root = tmpdir / "rst2pdf"
            assert luigi.build(
                [
                    TestWorkflow(
                        dataset_df=dataset_df_path,
                        result_path=str(root),
                        specifications_only=True,
                        report_config=report_config,
                    )
                ],
                local_scheduler=True,
            )

            assert (root / "TestWorkflow_specifications.pdf").exists()
            assert pdfdiff(
                root / "TestWorkflow_specifications.pdf",
                data_dir / "test_report_before_run" / "report_rst2pdf_with_config.pdf",
            )

        @pytest.mark.skipif(SKIP_IF_NO_LATEXMK, reason=REASON_NO_LATEXMK)
        def test_latexpdf_with_config(
            self, tmpdir, dataset_df_path, data_dir, TestWorkflow, report_config
        ):
            """Test PDF generation using latexpdf with a specific config."""
            root = tmpdir / "latexpdf"
            assert luigi.build(
                [
                    TestWorkflow(
                        dataset_df=dataset_df_path,
                        result_path=str(root),
                        report_type="latexpdf",
                        specifications_only=True,
                        report_config=report_config,
                    )
                ],
                local_scheduler=True,
            )

            assert (root / "TestWorkflow_specifications.pdf").exists()
            assert pdfdiff(
                root / "TestWorkflow_specifications.pdf",
                data_dir / "test_report_before_run" / "report_latexpdf_with_config.pdf",
            )

        def test_nested_workflows(
            self, tmpdir, dataset_df_path, data_dir, TestNestedWorkflow, report_config
        ):
            """Test PDF generation with nested workflows."""
            root = tmpdir / "rst2pdf_nested"
            assert luigi.build(
                [
                    TestNestedWorkflow(
                        dataset_df=dataset_df_path,
                        result_path=str(root),
                        specifications_only=True,
                        report_config=report_config,
                    )
                ],
                local_scheduler=True,
            )

            assert (root / "TestNestedWorkflow_specifications.pdf").exists()
            assert pdfdiff(
                root / "TestNestedWorkflow_specifications.pdf",
                data_dir / "test_report_before_run" / "report_rst2pdf_nested.pdf",
                threshold=25,
            )


class TestSkippableMixin:
    """Test the data_validation_framework.task.SkippableMixin class."""

    def test_fail_parent_type(self):
        """Test it can only be used with ElementValidationTask or SetValidationTask subclasses."""
        err_msg = (
            "The SkippableMixin can only be associated with children of ElementValidationTask"
            " or SetValidationTask"
        )

        class TestTask1(task.SkippableMixin(), luigi.Task):
            pass

        with pytest.raises(
            TypeError,
            match=err_msg,
        ):
            TestTask1()

        class TestTask2(task.SkippableMixin(), task.ValidationWorkflow):
            pass

        with pytest.raises(
            TypeError,
            match=err_msg,
        ):
            TestTask2()

    def test_skip_element_task(self, dataset_df_path, tmpdir):
        """Test that the task is properly skipped."""

        class TestSkippableTask(task.SkippableMixin(), task.ElementValidationTask):
            @staticmethod
            # pylint: disable=arguments-differ
            def validation_function(row, output_path, *args, **kwargs):
                if row["a"] <= 1:
                    return result.ValidationResult(is_valid=True)
                if row["a"] <= 2:
                    return result.ValidationResult(is_valid=False, comment="bad value")
                raise ValueError(f"Incorrect value {row['a']}")

        # Test with no given skip value (should be False by default)
        assert luigi.build(
            [
                TestSkippableTask(
                    dataset_df=dataset_df_path, result_path=str(tmpdir / "out_default")
                )
            ],
            local_scheduler=True,
        )

        report_data = pd.read_csv(tmpdir / "out_default" / "TestSkippableTask" / "report.csv")
        assert (report_data["is_valid"] == [True, False]).all()
        assert (report_data["comment"].isnull() == [True, False]).all()
        assert report_data.loc[1, "comment"] == "bad value"
        assert report_data["exception"].isnull().all()

        # Test with no given skip value (should be False by default)
        assert luigi.build(
            [
                TestSkippableTask(
                    dataset_df=dataset_df_path, result_path=str(tmpdir / "out_no_skip"), skip=False
                )
            ],
            local_scheduler=True,
        )

        report_data = pd.read_csv(tmpdir / "out_no_skip" / "TestSkippableTask" / "report.csv")
        assert (report_data["is_valid"] == [True, False]).all()
        assert (report_data["comment"].isnull() == [True, False]).all()
        assert report_data.loc[1, "comment"] == "bad value"
        assert report_data["exception"].isnull().all()

        # Test with no given skip value (should be False by default)
        assert luigi.build(
            [
                TestSkippableTask(
                    dataset_df=dataset_df_path, result_path=str(tmpdir / "out_skip"), skip=True
                )
            ],
            local_scheduler=True,
        )

        report_data = pd.read_csv(tmpdir / "out_skip" / "TestSkippableTask" / "report.csv")
        assert (
            report_data["is_valid"] == True  # noqa ; pylint: disable=singleton-comparison
        ).all()
        assert (report_data["comment"] == "Skipped by user.").all()
        assert report_data["exception"].isnull().all()

    def test_skip_set_task(self, dataset_df_path, tmpdir):
        """Test that the task is properly skipped."""

        class TestSkippableTask(task.SkippableMixin(), task.SetValidationTask):
            @staticmethod
            def validation_function(df, output_path, *args, **kwargs):
                # pylint: disable=no-member
                df["a"] *= 10
                df.loc[1, "is_valid"] = False
                df.loc[1, "ret_code"] = 1
                df[["a", "b"]].to_csv(output_path / "test.csv")

        # Test with no given skip value (should be False by default)
        assert luigi.build(
            [
                TestSkippableTask(
                    dataset_df=dataset_df_path, result_path=str(tmpdir / "out_default")
                )
            ],
            local_scheduler=True,
        )

        res = pd.read_csv(tmpdir / "out_default" / "TestSkippableTask" / "data" / "test.csv")
        expected = pd.read_csv(tmpdir / "dataset.csv")
        expected["a"] *= 10
        assert res.equals(expected)
        report_data = pd.read_csv(tmpdir / "out_default" / "TestSkippableTask" / "report.csv")
        assert (report_data["is_valid"] == [True, False]).all()
        assert report_data["comment"].isnull().all()
        assert report_data["exception"].isnull().all()

        # Test with skip = False
        assert luigi.build(
            [
                TestSkippableTask(
                    dataset_df=dataset_df_path, result_path=str(tmpdir / "out_no_skip"), skip=False
                )
            ],
            local_scheduler=True,
        )

        res = pd.read_csv(tmpdir / "out_no_skip" / "TestSkippableTask" / "data" / "test.csv")
        expected = pd.read_csv(tmpdir / "dataset.csv")
        expected["a"] *= 10
        assert res.equals(expected)
        report_data = pd.read_csv(tmpdir / "out_no_skip" / "TestSkippableTask" / "report.csv")
        assert (report_data["is_valid"] == [True, False]).all()
        assert report_data["comment"].isnull().all()
        assert report_data["exception"].isnull().all()

        # Test with skip = True
        assert luigi.build(
            [
                TestSkippableTask(
                    dataset_df=dataset_df_path, result_path=str(tmpdir / "out_skip"), skip=True
                )
            ],
            local_scheduler=True,
        )

        assert not (tmpdir / "out_skip" / "TestSkippableTask" / "data" / "test.csv").exists()
        report_data = pd.read_csv(tmpdir / "out_skip" / "TestSkippableTask" / "report.csv")
        assert (
            report_data["is_valid"] == True  # noqa ; pylint: disable=singleton-comparison
        ).all()
        assert (report_data["comment"] == "Skipped by user.").all()
        assert report_data["exception"].isnull().all()



"""Test the data_validation_framework.target module."""
# pylint: disable=missing-function-docstring
# pylint: disable=redefined-outer-name
# pylint: disable=unused-argument
import pytest

from data_validation_framework import target


@pytest.fixture
def reset_target_no_prefix():
    """Reset the default prefix of the OutputLocalTarget class."""
    target.OutputLocalTarget.set_default_prefix(None)
    yield
    target.OutputLocalTarget.set_default_prefix(None)


def test_ReportTarget(reset_target_no_prefix):
    """Test the ReportTarget class."""
    report_target = target.ReportTarget("a_path", create_parent=False)
    assert report_target.task_name is None
    assert report_target.path == "a_path"

    report_target_with_name = target.ReportTarget(
        "a_path", task_name="test_name", create_parent=False
    )
    assert report_target_with_name.task_name == "test_name"
    assert report_target_with_name.path == "a_path"


@pytest.fixture
def reset_target_relative_prefix():
    """Set the default prefix of the OutputLocalTarget class."""
    target.OutputLocalTarget.set_default_prefix("parent_default_prefix")
    yield
    target.OutputLocalTarget.set_default_prefix(None)


def test_ReportTarget_with_parent_prefix(reset_target_relative_prefix):
    """Test the ReportTarget class when a default prefix is given to TaggedOutputLocalTarget."""
    report_target = target.ReportTarget("a_path", create_parent=False)
    assert report_target.path == "parent_default_prefix/a_path"

    report_target = target.ReportTarget("a_path", prefix="report_prefix", create_parent=False)
    assert report_target.path == "parent_default_prefix/report_prefix/a_path"

    try:
        target.ReportTarget.set_default_prefix("report_default_prefix")

        report_target = target.ReportTarget("a_path", create_parent=False)
        assert report_target.path == "parent_default_prefix/report_default_prefix/a_path"

        report_target = target.ReportTarget("a_path", prefix="report_prefix", create_parent=False)
        assert report_target.path == "parent_default_prefix/report_prefix/a_path"
    finally:
        target.ReportTarget.set_default_prefix(None)

    report_target_with_name = target.ReportTarget(
        "a_path", task_name="test_name", create_parent=False
    )
    assert report_target_with_name.task_name == "test_name"



"""Util functions."""
import logging
import warnings
from copy import deepcopy

import pandas as pd

from data_validation_framework.util import report_missing_columns

L = logging.getLogger(__name__)


class ValidationResult(dict):
    """Class to represent a validation result as a dict.

    Some formatting is automatically executed during the dict creation:

    * if not given, the `ret_code` is set to `0` if `is_valid` is `True`, else it is set to `1`.
    * if the given `ret_code` is inconsistent with the given `is_valid` value, a ValueError is
        raised. The `ret_code` can have a value greater that 1 to mark a warning. In this case,
        the `is_valid` value can be either `True` or `False`.

    Args:
        is_valid (bool): Validation state.
        ret_code (int): The return code of the validation function.
        comment (str): A comment to explain why the validation failed.
        exception (str): The exception raised by the validation function if any.
    """

    def __init__(self, is_valid, ret_code=None, comment=None, exception=None, **kwargs):
        """ """
        if ret_code is None:
            if is_valid:
                ret_code = 0
            else:
                ret_code = 1
        else:
            if ret_code == 0 and not is_valid:
                raise ValueError("The 'is_valid' value must be True when 'ret_code' == 0.")
            if ret_code == 1 and is_valid:
                raise ValueError("The 'is_valid' value must be False when 'ret_code' == 1.")
            if ret_code not in [0, 1] and comment is None:
                warnings.warn("A comment should be set when the 'ret_code' is greater than 1.")

        super().__init__(
            is_valid=is_valid, ret_code=ret_code, comment=comment, exception=exception, **kwargs
        )


class ValidationResultSet(pd.DataFrame):
    """Class containing the results of a validation task.

    Args:
        df (pandas.DataFrame): Input DataFrame.
        output_columns (dict): New columns to save results in ({<column name>: <default value>})
    """

    out_cols = {
        "is_valid": True,  # if we don't yet have an is_valid, we consider them all
        "ret_code": None,
        "comment": None,
        "exception": None,
    }

    def __init__(self, *args, output_columns=None, **kwargs):
        super().__init__(*args, **kwargs)

        # Skip weird states due to Pandas and tqdm interactions
        data = kwargs.get("data", args[0] if len(args) > 0 else None)
        if not isinstance(
            data, (ValidationResultSet, pd.core.internals.managers.BlockManager)
        ) and not set(self.index).intersection(set(ValidationResultSet.out_cols.keys())):
            self.format_data(output_columns)

    @property
    def _constructor(self):
        """Pandas internal constructor property.

        See details here:
        https://pandas.pydata.org/pandas-docs/stable/development/extending.html#extending-subclassing-pandas
        """
        return ValidationResultSet

    @staticmethod
    def _col_comparator(col_index):
        """Comparator used to sort columns.

        The known columns are sorted in the given order, then the unknown columns are sorted
        at the end. In the unknown columns, the strings are sorted first then the other types.
        """
        out_cols = {k: f"   {num}" for num, k in enumerate(ValidationResultSet.out_cols.keys())}
        order = col_index.map(
            lambda i: out_cols.get(i, f"{'~' * (3 if isinstance(i, str) else 6)}_{i}")
        )
        return order

    def format_data(self, output_columns):
        """Format dataframe."""
        out_cols = deepcopy(ValidationResultSet.out_cols)

        # Add and erase task specific columns
        if output_columns is not None:
            out_cols.update(output_columns)

        for col, data in out_cols.items():
            if col not in self.columns:
                self[col] = data

        # Fill missing ret_code values
        self.loc[(self["ret_code"].isnull()) & (self["is_valid"]), "ret_code"] = 0
        self.loc[(self["ret_code"].isnull()) & (~self["is_valid"]), "ret_code"] = 1

        # Check missing columns in inputs
        req_cols = list(out_cols.keys())

        no_missing_cols, missing_report, _ = report_missing_columns(self, req_cols)

        if not no_missing_cols:  # pragma: no cover
            raise ValueError(missing_report)

        # Check column data
        for col in ["is_valid"]:
            assert not self[col].isnull().any(), f"The '{col}' column must not have null value."

        # Sort columns and return the DataFrame
        self.sort_index(axis=1, key=ValidationResultSet._col_comparator, inplace=True)



"""Util functions."""
import contextlib
import logging
import multiprocessing
import queue
import sys
import threading
import traceback
from itertools import repeat
from multiprocessing import Pool

import numpy as np
import pandas as pd
from tqdm import tqdm
from tqdm.contrib import DummyTqdmFile

L = logging.getLogger(__name__)


class StreamToQueue(DummyTqdmFile):
    """Fake file-like stream object that redirects all prints to a Queue."""

    def __init__(self, *args, message_queue=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.message_queue = message_queue

    def write(self, buf):
        """Redirect write calls to the Queue."""
        if len(buf.rstrip()) > 0:
            self.message_queue.put(buf)


@contextlib.contextmanager
def _std_out_err_redirect_func(*args, **kwargs):
    """Context manager used to redirect stdout and stderr to tqdm.write()."""
    orig_out_err = sys.stdout, sys.stderr
    try:
        if _std_out_err_redirect_func._redirect:  # pylint: disable=protected-access
            sys.stdout = StreamToQueue(sys.stdout, *args, **kwargs)
            sys.stderr = StreamToQueue(sys.stderr, *args, **kwargs)
        yield
    # Relay exceptions
    except Exception as exc:  # pragma: no cover
        raise exc
    # Always restore sys.stdout/err if necessary
    finally:
        sys.stdout, sys.stderr = orig_out_err


def _tqdm_wrapper(*args, **kwargs):
    # pylint: disable=protected-access
    _tqdm_wrapper._tqdm_queue.put(1)
    with _std_out_err_redirect_func(message_queue=_tqdm_wrapper._message_queue):
        res = try_operation(*args, **kwargs)
    return res


def _init_tqdm_global_queue(tqdm_queue, message_queue, redirect=True):
    _tqdm_wrapper._tqdm_queue = tqdm_queue  # pylint: disable=protected-access
    _tqdm_wrapper._message_queue = message_queue  # pylint: disable=protected-access
    _std_out_err_redirect_func._redirect = redirect  # pylint: disable=protected-access


def _reset_tqdm_global_queue():
    _tqdm_wrapper._tqdm_queue = None  # pylint: disable=protected-access
    _tqdm_wrapper._message_queue = None  # pylint: disable=protected-access
    _std_out_err_redirect_func._redirect = True  # pylint: disable=protected-access


def _apply_to_df_internal(data):
    (num, df), args, kwargs = data
    return num, df.apply(_tqdm_wrapper, axis=1, args=args, **kwargs)


def tqdm_worker(progress_bar, tqdm_queue):
    """Update progress bar using the Queue."""
    while True:
        db = tqdm_queue.get()
        if db is None:
            return
        progress_bar.update(db)


def message_worker(progress_bar, message_queue):
    """Write a message without interfering with the progress bar using the message Queue."""
    while True:
        message = message_queue.get()
        if message is None:
            return
        progress_bar.write(message)


def apply_to_df(df, func, *args, nb_processes=None, redirect_stdout=None, **kwargs):
    """Apply a function to df rows using tqdm."""
    if df.empty:
        return df
    nb_jobs = len(df)
    if redirect_stdout is None:
        redirect_stdout = True

    if nb_processes is None or nb_processes <= 1:
        # Serial computation
        is_parallel = False
        tqdm_queue = queue.Queue()
        message_queue = queue.Queue()
    else:
        # Parallel computation
        is_parallel = True
        nb_chunks = min(nb_jobs, nb_processes)
        chunks = enumerate(np.array_split(df, nb_chunks))
        tqdm_queue = multiprocessing.Queue()
        message_queue = multiprocessing.Queue()

    # Create the progress bar
    progress_bar = tqdm(total=nb_jobs, dynamic_ncols=True)
    tqdm_thread = threading.Thread(target=tqdm_worker, args=(progress_bar, tqdm_queue))
    message_thread = threading.Thread(target=message_worker, args=(progress_bar, message_queue))

    tqdm_thread.start()
    message_thread.start()

    if is_parallel:
        with Pool(
            nb_chunks,
            _init_tqdm_global_queue,
            [tqdm_queue, message_queue, redirect_stdout],
        ) as pool:

            # Start the computation
            results_list = pool.imap(
                _apply_to_df_internal,
                zip(chunks, repeat([func] + list(args)), repeat(kwargs)),
            )
            pool.close()

            # Wait for the last jobs to complete
            pool.join()

        # Gather all results
        all_res = pd.concat([j for i, j in sorted(results_list, key=lambda x: x[0])])

    else:
        _init_tqdm_global_queue(tqdm_queue, message_queue, redirect_stdout)
        _, all_res = _apply_to_df_internal(((0, df), [func] + list(args), kwargs))

    _reset_tqdm_global_queue()

    # Terminate the threads
    tqdm_queue.put_nowait(None)
    message_queue.put_nowait(None)
    tqdm_thread.join()
    message_thread.join()

    # Close the progress bar
    progress_bar.close()

    # Return the results
    return all_res


def try_operation(row, func, *args, **kwargs):
    """Try to apply a function on a :class:`pandas.Series`, and record exception."""
    if not row.is_valid:
        return row

    try:
        res = func(row, *args, **kwargs)
        row.loc[res.keys()] = list(res.values())
    except Exception:  # pylint: disable=broad-except
        exception = "".join(traceback.format_exception(*sys.exc_info()))
        L.warning("Exception for ID %s: %s", row.name, exception)
        row.exception = exception
        row.is_valid = False
        row.ret_code = 1
    return row


def check_missing_columns(df, required_columns):
    """Return a list of missing columns in a :class:`pandas.DataFrame`.

    Args:
        df (pandas.DataFrame): The DataFrame to check.
        required_columns (list): The list of column names. A column name can be a str for a one
            level column or either a list(tuple(str)) or a dict(list(str)) for a two-level column.
    """
    missing_cols = []
    for col in required_columns:  # pylint: disable=too-many-nested-blocks
        if isinstance(col, str) and col not in df.columns.get_level_values(0):
            missing_cols.append(col)
        elif isinstance(col, (list, tuple)) and col not in df.columns:
            missing_cols.append(col)
        elif isinstance(col, dict):
            for level1 in col:
                if isinstance(col[level1], str):
                    col_level2 = [col[level1]]
                else:
                    col_level2 = col[level1]
                if level1 not in df.columns.get_level_values(0):
                    for level2 in col_level2:
                        missing_cols.append((level1, level2))
                else:
                    for level2 in col_level2:
                        if level2 not in df[level1].columns.get_level_values(0):
                            missing_cols.append((level1, level2))
    return sorted(set(missing_cols))


def report_missing_columns(df, required_columns):
    """Check that required columns exist in a :class:`pandas.DataFrame`.

    Args:
        df (pandas.DataFrame): The DataFrame to check.
        required_columns (list): The list of column names. A column name can be a str for a one
            level column or either a list(tuple(str)) or a dict(list(str)) for a two-level column.
    """
    missing_cols = check_missing_columns(df, required_columns)
    report_str = f"Missing columns: {sorted(missing_cols)}"
    if missing_cols:
        return False, report_str, missing_cols
    return True, report_str, missing_cols



"""Test the data_validation_framework.util module."""
# pylint: disable=missing-function-docstring
import os
import re
import sys
import time

import pandas as pd
import pytest

from data_validation_framework import result
from data_validation_framework import util


def test_report_missing_columns():
    """Test that missing columns are properly retrieved."""
    df = pd.DataFrame({"a": [1, 2], "b": [3, 4]})
    res, report_str, missing_cols = util.report_missing_columns(df, ["a"])
    assert res
    assert report_str == "Missing columns: []"
    assert missing_cols == []

    res, report_str, missing_cols = util.report_missing_columns(df, ["a", "b"])
    assert res
    assert report_str == "Missing columns: []"
    assert missing_cols == []

    res, report_str, missing_cols = util.report_missing_columns(df, ["c"])
    assert not res
    assert report_str == "Missing columns: ['c']"
    assert missing_cols == ["c"]

    res, report_str, missing_cols = util.report_missing_columns(df, ["a", "b", "c"])
    assert not res
    assert report_str == "Missing columns: ['c']"
    assert missing_cols == ["c"]


def test_check_missing_columns():
    """Test that missing columns are properly checked."""
    df = pd.DataFrame({"a": [1, 2], "b": [3, 4], ("c", "d"): [5, 6]})

    assert util.check_missing_columns(df, ["a"]) == []
    assert util.check_missing_columns(df, ["c"]) == ["c"]
    assert util.check_missing_columns(df, [("c", "d")]) == []
    assert util.check_missing_columns(df, [("d", "c")]) == [("d", "c")]

    df.columns = pd.MultiIndex.from_tuples([("first_level", i) for i in df.columns.values])
    assert util.check_missing_columns(df, ["first_level"]) == []
    assert util.check_missing_columns(df, ["a"]) == ["a"]
    assert util.check_missing_columns(df, [{"first_level": ["a", "z"]}]) == [("first_level", "z")]
    assert util.check_missing_columns(df, [{"first_level": "a"}]) == []
    assert util.check_missing_columns(df, [{"unknown_first_level": "a"}]) == [
        ("unknown_first_level", "a")
    ]
    assert util.check_missing_columns(df, [{"unknown_first_level": ["a", "z"]}]) == [
        ("unknown_first_level", "a"),
        ("unknown_first_level", "z"),
    ]


def _tested_func(row, arg1, arg2):
    time.sleep(1)
    print("SOMETHING")
    print()
    if row["b"] == 6:
        raise ValueError("test 'b' value")
    return result.ValidationResult(
        True, new_data=f"{row['is_valid']}_{arg1}_{arg2}", pid=os.getpid()
    )


@pytest.mark.parametrize("nb_processes", [None, 1, 2, 100])
@pytest.mark.parametrize("redirect_stdout", [True, False])
def test_apply_to_df(nb_processes, redirect_stdout):
    """Test the apply_to_df function."""
    df = result.ValidationResultSet(
        pd.DataFrame(
            {
                "is_valid": [True, False, True, True],
                "a": [1, 2, 3, 4],
                "b": [4, 5, 6, 7],
                ("c", "d"): [7, 8, 9, 10],
            }
        ),
        output_columns={"new_data": None, "pid": None},
    )

    res = util.apply_to_df(
        df,
        _tested_func,
        "val1",
        "val2",
        nb_processes=nb_processes,
        redirect_stdout=redirect_stdout,
    )

    res_dict = res.to_dict()
    exception = res_dict.pop("exception")
    res_dict.pop("pid")
    assert res_dict == {
        "is_valid": {0: True, 1: False, 2: False, 3: True},
        "ret_code": {0: 0, 1: 1, 2: 1, 3: 0},
        "comment": {0: None, 1: None, 2: None, 3: None},
        "a": {0: 1, 1: 2, 2: 3, 3: 4},
        "b": {0: 4, 1: 5, 2: 6, 3: 7},
        "new_data": {0: "True_val1_val2", 1: None, 2: None, 3: "True_val1_val2"},
        ("c", "d"): {0: 7, 1: 8, 2: 9, 3: 10},
    }
    assert exception[0] is None
    assert exception[1] is None
    assert exception[3] is None

    assert res["pid"].isnull().tolist() == [False, True, True, False]
    if nb_processes is None or nb_processes == 1:
        assert res.loc[0, "pid"] >= 0
        assert res.loc[0, "pid"] == res.loc[3, "pid"]
    else:
        assert res.loc[0, "pid"] >= 0
        assert res.loc[3, "pid"] >= 0
        assert res.loc[0, "pid"] != res.loc[3, "pid"]
        assert len(res.loc[~res["pid"].isnull(), "pid"].unique()) == min(nb_processes, len(df) - 2)

    line_shift = sys.version_info.minor >= 3 and sys.version_info.minor >= 11
    exception_lines = exception[2].split("\n")
    assert exception_lines[0] == "Traceback (most recent call last):"
    assert exception_lines[2] == "    res = func(row, *args, **kwargs)"
    assert exception_lines[4 + line_shift] == "    raise ValueError(\"test 'b' value\")"
    assert exception_lines[5 + line_shift] == "ValueError: test 'b' value"
    assert (
        re.match(r"  File \"(\/.*?\.[\w:]+)\", line \d+, in try_operation", exception_lines[1])
        is not None
    )
    assert (
        re.match(
            r"  File \"(\/.*?\.[\w:]+)\", line \d+, in _tested_func",
            exception_lines[3 + line_shift],
        )
        is not None
    )

    # Test with an empty DF
    emty_res = util.apply_to_df(
        df.loc[[]],
        _tested_func,
        "val1",
        "val2",
    )

    pd.testing.assert_frame_equal(emty_res, df.loc[[]])



"""Test the data_validation_framework.result module."""
import pandas as pd
import pytest

from data_validation_framework import result


def test_ValidationResult():
    """Test the ValidationResult class."""
    assert result.ValidationResult(True)["ret_code"] == 0
    assert result.ValidationResult(True, 0)["ret_code"] == 0
    assert result.ValidationResult(False)["ret_code"] == 1
    assert result.ValidationResult(False, 1)["ret_code"] == 1
    assert result.ValidationResult(False, comment="test")["comment"] == "test"
    assert result.ValidationResult(False, exception="test")["exception"] == "test"
    assert result.ValidationResult(False, any_kwarg="test")["any_kwarg"] == "test"

    with pytest.raises(ValueError):
        result.ValidationResult(True, 1)

    with pytest.raises(ValueError):
        result.ValidationResult(False, 0)

    result.ValidationResult(True, 2, comment="test comment")

    with pytest.warns(
        UserWarning, match="A comment should be set when the 'ret_code' is greater than 1."
    ):
        result.ValidationResult(True, 2)


class TestValidationResultSet:
    """Test the ValidationResultSet class."""

    def test_defaults(self):
        """Check defaults."""
        df = result.ValidationResultSet(index=["a", "b"])
        assert df.to_dict() == {
            "is_valid": {"a": True, "b": True},
            "ret_code": {"a": 0, "b": 0},
            "comment": {"a": None, "b": None},
            "exception": {"a": None, "b": None},
        }

    def test_defaults_with_extra_data(self):
        """Check defaults with extra data."""
        df = result.ValidationResultSet({"a": [1, 2], "b": [3, 4]})
        assert df.to_dict() == {
            "is_valid": {0: True, 1: True},
            "ret_code": {0: 0, 1: 0},
            "comment": {0: None, 1: None},
            "exception": {0: None, 1: None},
            "a": {0: 1, 1: 2},
            "b": {0: 3, 1: 4},
        }

    def test_default_ret_code(self):
        """Check default return code."""
        df = result.ValidationResultSet({"is_valid": [True, False, True, False]})
        assert df.to_dict() == {
            "is_valid": {0: True, 1: False, 2: True, 3: False},
            "ret_code": {0: 0, 1: 1, 2: 0, 3: 1},
            "comment": {0: None, 1: None, 2: None, 3: None},
            "exception": {0: None, 1: None, 2: None, 3: None},
        }

    def test_output_columns(self):
        """Check output_columns argument."""
        df = result.ValidationResultSet(
            index=["a", "b"], output_columns={"col1": "val1", "col2": "val2"}
        )
        assert df.to_dict() == {
            "is_valid": {"a": True, "b": True},
            "ret_code": {"a": 0, "b": 0},
            "comment": {"a": None, "b": None},
            "exception": {"a": None, "b": None},
            "col1": {"a": "val1", "b": "val1"},
            "col2": {"a": "val2", "b": "val2"},
        }

    def test_output_columns_with_existing_columns(self):
        """Check output_columns argument with existing columns."""
        df = result.ValidationResultSet(
            {"is_valid": [True, False], "col1": ["test1", "test2"]},
            index=["a", "b"],
            output_columns={"col1": "val1", "col2": "val2"},
        )
        assert df.to_dict() == {
            "is_valid": {"a": True, "b": False},
            "ret_code": {"a": 0, "b": 1},
            "comment": {"a": None, "b": None},
            "exception": {"a": None, "b": None},
            "col1": {"a": "test1", "b": "test2"},
            "col2": {"a": "val2", "b": "val2"},
        }

    def test_constructor_df(self):
        """Check constructor with a pandas.DataFrame."""
        df = pd.DataFrame(
            {
                "is_valid": [True, False],
                "ret_code": [0, 1],
                "comment": [None, None],
                "exception": [None, None],
                "col1": ["test1", "test2"],
            },
            index=["a", "b"],
        )
        new_df = result.ValidationResultSet(df)
        assert new_df.equals(df)

    def test_constructor_ValidationResultSet(self):
        """Check constructor with a pandas.DataFrame."""
        df = result.ValidationResultSet(
            {"is_valid": [True, False], "col1": ["test1", "test2"]},
            index=["a", "b"],
            output_columns={"col1": "val1", "col2": "val2"},
        )
        new_df = result.ValidationResultSet(df)
        assert new_df.equals(df)

    @pytest.mark.filterwarnings("ignore::numpy.VisibleDeprecationWarning")
    def test_column_order(self):
        """Check column order."""
        df = result.ValidationResultSet(
            {"is_valid": [True, False], "col1": ["test1", "test2"]},
            index=["a", "b"],
            output_columns={"col1": "val1", "col2": "val2"},
        )
        df_order = result.ValidationResultSet(
            pd.DataFrame(df[["col2", "comment", "is_valid", "col1", "exception", "ret_code"]])
        )
        assert df_order.columns.tolist() == [
            "is_valid",
            "ret_code",
            "comment",
            "exception",
            "col1",
            "col2",
        ]

        df[("tuple", "in", "col")] = df["col1"]
        df_order = result.ValidationResultSet(
            pd.DataFrame(
                df[
                    [
                        "col2",
                        "comment",
                        "is_valid",
                        "col1",
                        ("tuple", "in", "col"),
                        "exception",
                        "ret_code",
                    ]
                ]
            )
        )
        assert df_order.columns.tolist() == [
            "is_valid",
            "ret_code",
            "comment",
            "exception",
            "col1",
            "col2",
            ("tuple", "in", "col"),
        ]

    def test_pandas_method(self):
        """Check using as a pandas.DataFrame."""
        df = result.ValidationResultSet(
            {"is_valid": [True, False], "col1": ["test1", "test2"]},
            index=["a", "b"],
            output_columns={"col1": "val1", "col2": "val2"},
        )
        new_df = result.ValidationResultSet(df)
        assert new_df.apply(lambda x: x.ret_code * 5, axis=1).to_dict() == {"a": 0, "b": 5}
        assert new_df["is_valid"].sum() == 1
        assert new_df["ret_code"].sum() == 1



"""Setup for the data-validation-framework package."""
from pathlib import Path

from setuptools import find_namespace_packages
from setuptools import setup

reqs = [
    "luigi>=3.1",
    "luigi-tools>=0.0.18",
    "pandas",
    "rst2pdf",
    "sphinx>=3,<6",
    "tqdm",
]
doc_reqs = [
    "m2r2",
    "sphinx",
    "sphinx-bluebrain-theme",
]
test_reqs = [
    "diff_pdf_visually>=1.6.2",
    "pause",
    "pytest",
    "pytest-cov",
    "pytest-html",
    "rst2pdf>=0.99",
]

setup(
    name="data-validation-framework",
    author="Blue Brain Project, EPFL",
    description="Simple framework to create data validation workflows.",
    long_description=Path("README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    url="https://data-validation-framework.readthedocs.io",
    project_urls={
        "Tracker": "https://github.com/BlueBrain/data-validation-framework/issues",
        "Source": "https://github.com/BlueBrain/data-validation-framework",
    },
    license="Apache License 2.0",
    packages=find_namespace_packages(include=["data_validation_framework*"]),
    python_requires=">=3.8",
    use_scm_version=True,
    setup_requires=[
        "setuptools_scm",
    ],
    install_requires=reqs,
    extras_require={
        "docs": doc_reqs,
        "test": test_reqs,
    },
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)
